# dynamiccolor/__init__.py

# --- color_spec_delegate ---
from PyMCUlib.dynamiccolor.color_spec_delegate import (
    ColorSpecDelegate,
    SpecVersion,
)

# --- color_spec ---
from PyMCUlib.dynamiccolor.color_spec import (
    get_spec,
)

# --- contrast_curve ---
from PyMCUlib.dynamiccolor.contrast_curve import (
    ContrastCurve,
)

# --- dynamic_color ---
from PyMCUlib.dynamiccolor.dynamic_color import (
    DynamicColor,
    extend_spec_version,
    validate_extended_color,
)

# --- dynamic_scheme ---
from PyMCUlib.dynamiccolor.dynamic_scheme import (
    DynamicScheme,
)

# --- material_dynamic_colors ---
from PyMCUlib.dynamiccolor.material_dynamic_colors import (
    MaterialDynamicColors,
)

# --- tone_delta_pair ---
from PyMCUlib.dynamiccolor.tone_delta_pair import (
    ToneDeltaPair,
    DeltaConstraint,
    TonePolarity,
)

# --- variant ---
from PyMCUlib.dynamiccolor.variant import (
    Variant,
)

__all__ = [
    # color_spec_delegate
    "ColorSpecDelegate",
    "SpecVersion",
    # color_spec
    "get_spec",
    # contrast_curve
    "ContrastCurve",
    # dynamic_color
    "DynamicColor",
    "extend_spec_version",
    "validate_extended_color",
    # dynamic_scheme
    "DynamicScheme",
    # material_dynamic_colors
    "MaterialDynamicColors",
    # tone_delta_pair
    "ToneDeltaPair",
    "DeltaConstraint",
    "TonePolarity",
    # variant
    "Variant",
]


# palettes/core_palettes.py

"""
Comprises foundational palettes to build a color scheme. Generated from a
source color, these palettes will then be part of a DynamicScheme together
with appearance preferences.
"""

from PyMCUlib.palettes.tonal_palette import TonalPalette


class CorePalettes:
    """
    Comprises foundational palettes to build a color scheme. Generated from a
    source color, these palettes will then be part of a DynamicScheme together
    with appearance preferences.
    """

    def __init__(
            self,
            primary: TonalPalette,
            secondary: TonalPalette,
            tertiary: TonalPalette,
            neutral: TonalPalette,
            neutral_variant: TonalPalette) -> None:
        """
        Initialize a CorePalettes with tonally-consistent palettes.
        
        Args:
            primary: TonalPalette for primary color
            secondary: TonalPalette for secondary color
            tertiary: TonalPalette for tertiary color
            neutral: TonalPalette for neutral color
            neutral_variant: TonalPalette for neutral variant color
        """
        self.primary = primary
        self.secondary = secondary
        self.tertiary = tertiary
        self.neutral = neutral
        self.neutral_variant = neutral_variant


# quantize/__init__.py

# --- point_provider ---
from PyMCUlib.quantize.point_provider import PointProvider

# --- lab_point_provider ---
from PyMCUlib.quantize.lab_point_provider import LabPointProvider

# --- quantizer_map ---
from PyMCUlib.quantize.quantizer_map import QuantizerMap

# --- quantizer_wu ---
from PyMCUlib.quantize.quantizer_wu import QuantizerWu

# --- quantizer_wsmeans ---
from PyMCUlib.quantize.quantizer_wsmeans import QuantizerWsmeans

# --- quantizer_celebi ---
from PyMCUlib.quantize.quantizer_celebi import QuantizerCelebi

__all__ = [
    # interface and abstract class
    "PointProvider",
    # Lab space implementation
    "LabPointProvider",
    # basic mapping quantization
    "QuantizerMap",
    # Wu algorithm
    "QuantizerWu",
    # WSMeans algorithm
    "QuantizerWsmeans",
    # Celebi algorithm (Wu + WSMeans)
    "QuantizerCelebi",
]


# utils/__init__.py

# --- color_utils ---
from PyMCUlib.utils.color_utils import (
    argb_from_rgb,
    argb_from_linrgb,
    alpha_from_argb,
    red_from_argb,
    green_from_argb,
    blue_from_argb,
    is_opaque,
    argb_from_xyz,
    xyz_from_argb,
    argb_from_lab,
    lab_from_argb,
    argb_from_lstar,
    lstar_from_argb,
    y_from_lstar,
    lstar_from_y,
    linearized,
    delinearized,
    white_point_d65,
)

# --- string_utils ---
from PyMCUlib.utils.string_utils import (
    hex_from_argb,
    argb_from_hex,
)

# --- image_utils ---
from PyMCUlib.utils.image_utils import (
    source_color_from_image,
    source_color_from_image_bytes,
    source_color_from_file,
)

# --- math_utils ---
from PyMCUlib.utils.math_utils import (
    signum,
    lerp,
    clamp_int,
    clamp_double,
    sanitize_degrees_int,
    sanitize_degrees_double,
    rotation_direction,
    difference_degrees,
    matrix_multiply,
)

__all__ = [
    # color_utils
    "argb_from_rgb",
    "argb_from_linrgb",
    "alpha_from_argb",
    "red_from_argb",
    "green_from_argb",
    "blue_from_argb",
    "is_opaque",
    "argb_from_xyz",
    "xyz_from_argb",
    "argb_from_lab",
    "lab_from_argb",
    "argb_from_lstar",
    "lstar_from_argb",
    "y_from_lstar",
    "lstar_from_y",
    "linearized",
    "delinearized",
    "white_point_d65",
    # string_utils
    "hex_from_argb",
    "argb_from_hex",
    # image_utils
    "source_color_from_image",
    "source_color_from_image_bytes",
    "source_color_from_file",
    # math_utils
    "signum",
    "lerp",
    "clamp_int",
    "clamp_double",
    "sanitize_degrees_int",
    "sanitize_degrees_double",
    "rotation_direction",
    "difference_degrees",
    "matrix_multiply",
    # theme_utils
    "CustomColor",
    "ColorGroup",
    "CustomColorGroup",
    "Theme",
    "ApplyThemeOptions",
    "theme_from_source_color",
    "theme_from_image",
    "custom_color",
    "apply_theme",
    "set_scheme_properties",
]

def theme_from_source_color(source: int, custom_colors=None):
    from PyMCUlib.utils.theme_utils import theme_from_source_color as _f
    return _f(source, custom_colors)

def theme_from_image(image, custom_colors=None):
    from PyMCUlib.utils.theme_utils import theme_from_image as _f
    return _f(image, custom_colors)

def custom_color(source: int, color):
    from PyMCUlib.utils.theme_utils import custom_color as _f
    return _f(source, color)

def apply_theme(theme, options=None):
    from PyMCUlib.utils.theme_utils import apply_theme as _f
    return _f(theme, options)

def set_scheme_properties(css_vars, scheme, suffix=""):
    from PyMCUlib.utils.theme_utils import set_scheme_properties as _f
    return _f(css_vars, scheme, suffix)

def __getattr__(name):
    if name == "CustomColor":
        from PyMCUlib.utils.theme_utils import CustomColor
        return CustomColor
    if name == "ColorGroup":
        from PyMCUlib.utils.theme_utils import ColorGroup
        return ColorGroup
    if name == "CustomColorGroup":
        from PyMCUlib.utils.theme_utils import CustomColorGroup
        return CustomColorGroup
    if name == "Theme":
        from PyMCUlib.utils.theme_utils import Theme
        return Theme
    if name == "ApplyThemeOptions":
        from PyMCUlib.utils.theme_utils import ApplyThemeOptions
        return ApplyThemeOptions
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")

def __dir__():
    return sorted(list(globals().keys()) + __all__)


# contrast/contrast.py

from PyMCUlib.utils import color_utils
from PyMCUlib.utils import math_utils

class Contrast:
    """
    Utility methods for calculating contrast given two colors, or calculating a
    color given one color and a contrast ratio.

    Contrast ratio is calculated using XYZ's Y. When linearized to match human
    perception, Y becomes HCT's tone and L*a*b*'s' L*. Informally, this is the
    lightness of a color.

    Methods refer to tone, T in the the HCT color space.
    Tone is equivalent to L* in the L*a*b* color space, or L in the LCH color
    space.
    """

    @staticmethod
    def ratio_of_tones(tone_a: float, tone_b: float) -> float:
        """
        Returns a contrast ratio, which ranges from 1 to 21.

        Args:
            tone_a: Tone between 0 and 100. Values outside will be clamped.
            tone_b: Tone between 0 and 100. Values outside will be clamped.

        Returns:
            Contrast ratio between the two tones
        """
        tone_a = math_utils.clamp_double(0.0, 100.0, tone_a)
        tone_b = math_utils.clamp_double(0.0, 100.0, tone_b)
        return Contrast.ratio_of_ys(color_utils.y_from_lstar(tone_a), color_utils.y_from_lstar(tone_b))

    @staticmethod
    def ratio_of_ys(y1: float, y2: float) -> float:
        """
        Calculates the contrast ratio between two luminance values.

        Args:
            y1: First luminance value
            y2: Second luminance value

        Returns:
            Contrast ratio between the two luminance values
        """
        lighter = y1 if y1 > y2 else y2
        darker = y1 if lighter == y2 else y2
        return (lighter + 5.0) / (darker + 5.0)

    @staticmethod
    def lighter(tone: float, ratio: float) -> float:
        """
        Returns a tone >= tone parameter that ensures ratio parameter.
        Return value is between 0 and 100.
        Returns -1 if ratio cannot be achieved with tone parameter.

        Args:
            tone: Tone return value must contrast with.
                Range is 0 to 100. Invalid values will result in -1 being returned.
            ratio: Contrast ratio of return value and tone.
                Range is 1 to 21, invalid values have undefined behavior.

        Returns:
            Tone value that achieves the desired contrast ratio, or -1 if not possible
        """
        if tone < 0.0 or tone > 100.0:
            return -1.0

        dark_y = color_utils.y_from_lstar(tone)
        light_y = ratio * (dark_y + 5.0) - 5.0
        real_contrast = Contrast.ratio_of_ys(light_y, dark_y)
        delta = abs(real_contrast - ratio)
        if real_contrast < ratio and delta > 0.04:
            return -1

        # Ensure gamut mapping, which requires a 'range' on tone, will still result
        # the correct ratio by darkening slightly.
        return_value = color_utils.lstar_from_y(light_y) + 0.4
        if return_value < 0 or return_value > 100:
            return -1
        return return_value

    @staticmethod
    def darker(tone: float, ratio: float) -> float:
        """
        Returns a tone <= tone parameter that ensures ratio parameter.
        Return value is between 0 and 100.
        Returns -1 if ratio cannot be achieved with tone parameter.

        Args:
            tone: Tone return value must contrast with.
                Range is 0 to 100. Invalid values will result in -1 being returned.
            ratio: Contrast ratio of return value and tone.
                Range is 1 to 21, invalid values have undefined behavior.

        Returns:
            Tone value that achieves the desired contrast ratio, or -1 if not possible
        """
        if tone < 0.0 or tone > 100.0:
            return -1.0

        light_y = color_utils.y_from_lstar(tone)
        dark_y = ((light_y + 5.0) / ratio) - 5.0
        real_contrast = Contrast.ratio_of_ys(light_y, dark_y)

        delta = abs(real_contrast - ratio)
        if real_contrast < ratio and delta > 0.04:
            return -1

        # Ensure gamut mapping, which requires a 'range' on tone, will still result
        # the correct ratio by darkening slightly.
        return_value = color_utils.lstar_from_y(dark_y) - 0.4
        if return_value < 0 or return_value > 100:
            return -1
        return return_value

    @staticmethod
    def lighter_unsafe(tone: float, ratio: float) -> float:
        """
        Returns a tone >= tone parameter that ensures ratio parameter.
        Return value is between 0 and 100.
        Returns 100 if ratio cannot be achieved with tone parameter.

        This method is unsafe because the returned value is guaranteed to be in
        bounds for tone, i.e. between 0 and 100. However, that value may not reach
        the ratio with tone. For example, there is no color lighter than T100.

        Args:
            tone: Tone return value must contrast with.
                Range is 0 to 100. Invalid values will result in 100 being returned.
            ratio: Desired contrast ratio of return value and tone parameter.
                Range is 1 to 21, invalid values have undefined behavior.

        Returns:
            Tone value that achieves the desired contrast ratio, or 100 if not possible
        """
        lighter_safe = Contrast.lighter(tone, ratio)
        return 100.0 if lighter_safe < 0.0 else lighter_safe

    @staticmethod
    def darker_unsafe(tone: float, ratio: float) -> float:
        """
        Returns a tone <= tone parameter that ensures ratio parameter.
        Return value is between 0 and 100.
        Returns 0 if ratio cannot be achieved with tone parameter.

        This method is unsafe because the returned value is guaranteed to be in
        bounds for tone, i.e. between 0 and 100. However, that value may not reach
        the ratio with tone. For example, there is no color darker than T0.

        Args:
            tone: Tone return value must contrast with.
                Range is 0 to 100. Invalid values will result in 0 being returned.
            ratio: Desired contrast ratio of return value and tone parameter.
                Range is 1 to 21, invalid values have undefined behavior.

        Returns:
            Tone value that achieves the desired contrast ratio, or 0 if not possible
        """
        darker_safe = Contrast.darker(tone, ratio)
        return 0.0 if darker_safe < 0.0 else darker_safe


# dynamiccolor/dynamic_color.py

from __future__ import annotations
from typing import Callable, Dict, Optional, Protocol, TypedDict, TYPE_CHECKING
import math

if TYPE_CHECKING:
    from .dynamic_scheme import DynamicScheme

from PyMCUlib.dynamiccolor.color_spec_delegate import SpecVersion
from PyMCUlib.dynamiccolor.contrast_curve import ContrastCurve
from PyMCUlib.dynamiccolor.tone_delta_pair import ToneDeltaPair
from PyMCUlib.contrast.contrast import Contrast
from PyMCUlib.hct.hct import Hct
from PyMCUlib.palettes.tonal_palette import TonalPalette
from PyMCUlib.utils import math_utils


class FromPaletteOptions(TypedDict, total=False):
    """
    Options for creating a DynamicColor from a palette.
    
    Args:
        name: The name of the dynamic color. Defaults to empty.
        palette: Function that provides a TonalPalette given DynamicScheme.
        tone: Function that provides a tone given DynamicScheme.
        chroma_multiplier: A factor that multiplies the chroma for this color.
        is_background: Whether this dynamic color is a background, with some other color as the foreground.
        background: The background of the dynamic color (as a function of a DynamicScheme), if it exists.
        second_background: A second background of the dynamic color (as a function of a DynamicScheme), if it exists.
        contrast_curve: A ContrastCurve object specifying how its contrast against its background should behave.
        tone_delta_pair: A ToneDeltaPair object specifying a tone delta constraint between two colors.
    """
    name: str
    palette: Callable[["DynamicScheme"], TonalPalette]
    tone: Callable[["DynamicScheme"], float]
    chroma_multiplier: Callable[["DynamicScheme"], float]
    is_background: bool
    background: Callable[["DynamicScheme"], Optional[DynamicColor]]
    second_background: Callable[["DynamicScheme"], Optional[DynamicColor]]
    contrast_curve: Callable[["DynamicScheme"], Optional[ContrastCurve]]
    tone_delta_pair: Callable[["DynamicScheme"], Optional[ToneDeltaPair]]


class ColorCalculationDelegate(Protocol):
    """
    A delegate that provides the HCT and tone of a DynamicColor.

    This is used to allow different implementations of the color calculation
    logic for different spec versions.
    """
    def get_hct(self, scheme: "DynamicScheme", color: DynamicColor) -> Hct:
        ...

    def get_tone(self, scheme: "DynamicScheme", color: DynamicColor) -> float:
        ...


def validate_extended_color(
        original_color: DynamicColor, spec_version: SpecVersion,
        extended_color: DynamicColor) -> None:
    """
    Validates that an extended color is compatible with the original color.
    
    Args:
        original_color: The original color.
        spec_version: The spec version to extend.
        extended_color: The color with the values to extend.
        
    Raises:
        ValueError: If the extended color is not compatible with the original color.
    """
    if original_color.name != extended_color.name:
        raise ValueError(
            f"Attempting to extend color {original_color.name} with color "
            f"{extended_color.name} of different name for spec version {spec_version}.")
    
    if original_color.is_background != extended_color.is_background:
        raise ValueError(
            f"Attempting to extend color {original_color.name} as a "
            f"{'background' if original_color.is_background else 'foreground'} with color "
            f"{extended_color.name} as a "
            f"{'background' if extended_color.is_background else 'foreground'} "
            f"for spec version {spec_version}.")


def extend_spec_version(
        original_color: DynamicColor, spec_version: SpecVersion,
        extended_color: DynamicColor) -> DynamicColor:
    """
    Returns a new DynamicColor that is the same as the original color, but with
    the extended dynamic color's constraints for the given spec version.

    Args:
        original_color: The original color.
        spec_version: The spec version to extend.
        extended_color: The color with the values to extend.
        
    Returns:
        A new DynamicColor that extends the original color.
    """
    validate_extended_color(original_color, spec_version, extended_color)

    return DynamicColor.from_palette({
        'name': original_color.name,
        'palette': lambda s: extended_color.palette(s) if s.spec_version == spec_version else original_color.palette(s),
        'tone': lambda s: extended_color.tone(s) if s.spec_version == spec_version else original_color.tone(s),
        'is_background': original_color.is_background,
        'chroma_multiplier': lambda s: (
            (extended_color.chroma_multiplier(s) if extended_color.chroma_multiplier is not None else 1)
            if s.spec_version == spec_version else
            (original_color.chroma_multiplier(s) if original_color.chroma_multiplier is not None else 1)
        ),
        'background': lambda s: (
            (extended_color.background(s) if extended_color.background is not None else None)
            if s.spec_version == spec_version else
            (original_color.background(s) if original_color.background is not None else None)
        ),
        'second_background': lambda s: (
            (extended_color.second_background(s) if extended_color.second_background is not None else None)
            if s.spec_version == spec_version else
            (original_color.second_background(s) if original_color.second_background is not None else None)
        ),
        'contrast_curve': lambda s: (
            (extended_color.contrast_curve(s) if extended_color.contrast_curve is not None else None)
            if s.spec_version == spec_version else
            (original_color.contrast_curve(s) if original_color.contrast_curve is not None else None)
        ),
        'tone_delta_pair': lambda s: (
            (extended_color.tone_delta_pair(s) if extended_color.tone_delta_pair is not None else None)
            if s.spec_version == spec_version else
            (original_color.tone_delta_pair(s) if original_color.tone_delta_pair is not None else None)
        ),
    })


class DynamicColor:
    """
    A color that adjusts itself based on UI state provided by DynamicScheme.

    Colors without backgrounds do not change tone when contrast changes. Colors
    with backgrounds become closer to their background as contrast lowers, and
    further when contrast increases.

    Prefer static constructors. They require either a hexcode, a palette and
    tone, or a hue and chroma. Optionally, they can provide a background
    DynamicColor.
    """

    @staticmethod
    def from_palette(args: FromPaletteOptions) -> DynamicColor:
        """
        Create a DynamicColor defined by a TonalPalette and HCT tone.

        Args:
            args: Functions with DynamicScheme as input. Must provide a palette
                and tone. May provide a background DynamicColor and ToneDeltaPair.
                
        Returns:
            A new DynamicColor instance.
        """
        return DynamicColor(
            args.get('name', ''),
            args['palette'],
            args.get('tone', DynamicColor.get_initial_tone_from_background(args.get('background'))),
            args.get('is_background', False),
            args.get('chroma_multiplier'),
            args.get('background'),
            args.get('second_background'),
            args.get('contrast_curve'),
            args.get('tone_delta_pair'),
        )

    @staticmethod
    def get_initial_tone_from_background(
            background: Optional[Callable[["DynamicScheme"], Optional[DynamicColor]]]) -> Callable[["DynamicScheme"], float]:
        """
        Gets the initial tone from a background function.
        
        Args:
            background: A function that returns a background color given a DynamicScheme.
            
        Returns:
            A function that returns the initial tone.
        """
        if background is None:
            return lambda s: 50.0
        
        return lambda s: background(s).get_tone(s) if background(s) else 50.0

    def __init__(
            self,
            name: str,
            palette: Callable[["DynamicScheme"], TonalPalette],
            tone: Callable[["DynamicScheme"], float],
            is_background: bool,
            chroma_multiplier: Optional[Callable[["DynamicScheme"], float]] = None,
            background: Optional[Callable[["DynamicScheme"], Optional[DynamicColor]]] = None,
            second_background: Optional[Callable[["DynamicScheme"], Optional[DynamicColor]]] = None,
            contrast_curve: Optional[Callable[["DynamicScheme"], Optional[ContrastCurve]]] = None,
            tone_delta_pair: Optional[Callable[["DynamicScheme"], Optional[ToneDeltaPair]]] = None,
    ):
        """
        The base constructor for DynamicColor.

        Strongly prefer using one of the convenience constructors. This class is
        arguably too flexible to ensure it can support any scenario. Functional
        arguments allow overriding without risks that come with subclasses.

        For example, the default behavior of adjust tone at max contrast
        to be at a 7.0 ratio with its background is principled and
        matches accessibility guidance. That does not mean it's the desired
        approach for every design system, and every color pairing,
        always, in every case.

        Args:
            name: The name of the dynamic color. Defaults to empty.
            palette: Function that provides a TonalPalette given DynamicScheme.
            tone: Function that provides a tone, given a DynamicScheme.
            is_background: Whether this dynamic color is a background, with some
                other color as the foreground.
            chroma_multiplier: A factor that multiplies the chroma for this color.
            background: The background of the dynamic color (as a function of a
                DynamicScheme), if it exists.
            second_background: A second background of the dynamic color (as a
                function of a DynamicScheme), if it exists.
            contrast_curve: A ContrastCurve object specifying how its contrast
                against its background should behave.
            tone_delta_pair: A ToneDeltaPair object specifying a tone delta
                constraint between two colors.
        """
        self.name = name
        self.palette = palette
        self.tone = tone
        self.is_background = is_background
        self.chroma_multiplier = chroma_multiplier
        self.background = background
        self.second_background = second_background
        self.contrast_curve = contrast_curve
        self.tone_delta_pair = tone_delta_pair
        self.hct_cache: Dict["DynamicScheme", Hct] = {}
        
        if (not background) and second_background:
            raise ValueError(
                f"Color {name} has second_background defined, but background is not defined.")
        
        if (not background) and contrast_curve:
            raise ValueError(
                f"Color {name} has contrast_curve defined, but background is not defined.")
        
        if background and not contrast_curve:
            raise ValueError(
                f"Color {name} has background defined, but contrast_curve is not defined.")

    def clone(self) -> DynamicColor:
        """
        Returns a deep copy of this DynamicColor.
        
        Returns:
            A new DynamicColor instance with the same properties.
        """
        return DynamicColor.from_palette({
            'name': self.name,
            'palette': self.palette,
            'tone': self.tone,
            'is_background': self.is_background,
            'chroma_multiplier': self.chroma_multiplier,
            'background': self.background,
            'second_background': self.second_background,
            'contrast_curve': self.contrast_curve,
            'tone_delta_pair': self.tone_delta_pair,
        })

    def clear_cache(self) -> None:
        """
        Clears the cache of HCT values for this color. For testing or debugging
        purposes.
        """
        self.hct_cache.clear()

    def get_argb(self, scheme: "DynamicScheme") -> int:
        """
        Returns a ARGB integer (i.e. a hex code).

        Args:
            scheme: Defines the conditions of the user interface, for example,
                whether or not it is dark mode or light mode, and what the desired
                contrast level is.
                
        Returns:
            An ARGB color as an integer.
        """
        return self.get_hct(scheme).to_int()

    def get_hct(self, scheme: "DynamicScheme") -> Hct:
        """
        Returns a color, expressed in the HCT color space, that this
        DynamicColor is under the conditions in scheme.

        Args:
            scheme: Defines the conditions of the user interface, for example,
                whether or not it is dark mode or light mode, and what the desired
                contrast level is.
                
        Returns:
            An HCT color.
        """
        cached_answer = self.hct_cache.get(scheme)
        if cached_answer is not None:
            return cached_answer
        
        answer = get_spec(scheme.spec_version).get_hct(scheme, self)
        
        if len(self.hct_cache) > 4:
            self.hct_cache.clear()
        
        self.hct_cache[scheme] = answer
        return answer

    def get_tone(self, scheme: "DynamicScheme") -> float:
        """
        Returns a tone, T in the HCT color space, that this DynamicColor is under
        the conditions in scheme.

        Args:
            scheme: Defines the conditions of the user interface, for example,
                whether or not it is dark mode or light mode, and what the desired
                contrast level is.
                
        Returns:
            A tone value between 0 and 100.
        """
        return get_spec(scheme.spec_version).get_tone(scheme, self)

    @staticmethod
    def foreground_tone(bg_tone: float, ratio: float) -> float:
        """
        Given a background tone, finds a foreground tone, while ensuring they reach
        a contrast ratio that is as close to [ratio] as possible.

        Args:
            bg_tone: Tone in HCT. Range is 0 to 100, undefined behavior when it
                falls outside that range.
            ratio: The contrast ratio desired between bgTone and the return
                value.
                
        Returns:
            A tone value between 0 and 100.
        """
        lighter_tone = Contrast.lighter_unsafe(bg_tone, ratio)
        darker_tone = Contrast.darker_unsafe(bg_tone, ratio)
        lighter_ratio = Contrast.ratio_of_tones(lighter_tone, bg_tone)
        darker_ratio = Contrast.ratio_of_tones(darker_tone, bg_tone)
        prefer_lighter = DynamicColor.tone_prefers_light_foreground(bg_tone)

        if prefer_lighter:
            # This handles an edge case where the initial contrast ratio is high
            # (ex. 13.0), and the ratio passed to the function is that high
            # ratio, and both the lighter and darker ratio fails to pass that
            # ratio.
            #
            # This was observed with Tonal Spot's On Primary Container turning
            # black momentarily between high and max contrast in light mode. PC's
            # standard tone was T90, OPC's was T10, it was light mode, and the
            # contrast value was 0.6568521221032331.
            negligible_difference = (abs(lighter_ratio - darker_ratio) < 0.1 and
                                    lighter_ratio < ratio and darker_ratio < ratio)
            return (lighter_tone if (lighter_ratio >= ratio or 
                                    lighter_ratio >= darker_ratio or 
                                    negligible_difference) 
                    else darker_tone)
        else:
            return darker_tone if (darker_ratio >= ratio or darker_ratio >= lighter_ratio) else lighter_tone

    @staticmethod
    def tone_prefers_light_foreground(tone: float) -> bool:
        """
        Returns whether [tone] prefers a light foreground.

        People prefer white foregrounds on ~T60-70. Observed over time, and also
        by Andrew Somers during research for APCA.

        T60 used as to create the smallest discontinuity possible when skipping
        down to T49 in order to ensure light foregrounds.
        Since `tertiaryContainer` in dark monochrome scheme requires a tone of
        60, it should not be adjusted. Therefore, 60 is excluded here.
        
        Args:
            tone: A tone value between 0 and 100.
            
        Returns:
            True if the tone prefers a light foreground, False otherwise.
        """
        return math.floor(tone + 0.5) < 60.0

    @staticmethod
    def tone_allows_light_foreground(tone: float) -> bool:
        """
        Returns whether [tone] can reach a contrast ratio of 4.5 with a lighter
        color.
        
        Args:
            tone: A tone value between 0 and 100.
            
        Returns:
            True if the tone allows a light foreground, False otherwise.
        """
        return math.floor(tone + 0.5) <= 49.0

    @staticmethod
    def enable_light_foreground(tone: float) -> float:
        """
        Adjusts a tone such that white has 4.5 contrast, if the tone is
        reasonably close to supporting it.
        
        Args:
            tone: A tone value between 0 and 100.
            
        Returns:
            An adjusted tone value between 0 and 100.
        """
        if (DynamicColor.tone_prefers_light_foreground(tone) and
                not DynamicColor.tone_allows_light_foreground(tone)):
            return 49.0
        return tone


class ColorCalculationDelegateImpl2021(ColorCalculationDelegate):
    """
    A delegate for the color calculation of a DynamicScheme in the 2021 spec.
    """

    def get_hct(self, scheme: "DynamicScheme", color: DynamicColor) -> Hct:
        """
        Get the HCT color for the given scheme and color.
        
        Args:
            scheme: The dynamic scheme.
            color: The dynamic color.
            
        Returns:
            An HCT color.
        """
        tone = color.get_tone(scheme)
        palette = color.palette(scheme)
        return palette.get_hct(tone)

    def get_tone(self, scheme: "DynamicScheme", color: DynamicColor) -> float:
        """
        Get the tone for the given scheme and color.
        
        Args:
            scheme: The dynamic scheme.
            color: The dynamic color.
            
        Returns:
            A tone value between 0 and 100.
        """
        decreasing_contrast = scheme.contrast_level < 0
        tone_delta_pair = color.tone_delta_pair(scheme) if color.tone_delta_pair else None

        # Case 1: dual foreground, pair of colors with delta constraint.
        if tone_delta_pair:
            role_a = tone_delta_pair.role_a
            role_b = tone_delta_pair.role_b
            delta = tone_delta_pair.delta
            polarity = tone_delta_pair.polarity
            stay_together = tone_delta_pair.stay_together

            a_is_nearer = (polarity == 'nearer' or
                          (polarity == 'lighter' and not scheme.is_dark) or
                          (polarity == 'darker' and scheme.is_dark))
            nearer = role_a if a_is_nearer else role_b
            farther = role_b if a_is_nearer else role_a
            am_nearer = color.name == nearer.name
            expansion_dir = 1 if scheme.is_dark else -1
            n_tone = nearer.tone(scheme)
            f_tone = farther.tone(scheme)

            # 1st round: solve to min for each, if background and contrast curve
            # are defined.
            if color.background and nearer.contrast_curve and farther.contrast_curve:
                bg = color.background(scheme)
                n_contrast_curve = nearer.contrast_curve(scheme)
                f_contrast_curve = farther.contrast_curve(scheme)
                if bg and n_contrast_curve and f_contrast_curve:
                    bg_tone = bg.get_tone(scheme)
                    n_contrast = n_contrast_curve.get(scheme.contrast_level)
                    f_contrast = f_contrast_curve.get(scheme.contrast_level)
                    # If a color is good enough, it is not adjusted.
                    # Initial and adjusted tones for `nearer`
                    if Contrast.ratio_of_tones(bg_tone, n_tone) < n_contrast:
                        n_tone = DynamicColor.foreground_tone(bg_tone, n_contrast)
                    # Initial and adjusted tones for `farther`
                    if Contrast.ratio_of_tones(bg_tone, f_tone) < f_contrast:
                        f_tone = DynamicColor.foreground_tone(bg_tone, f_contrast)
                    if decreasing_contrast:
                        # If decreasing contrast, adjust color to the "bare minimum"
                        # that satisfies contrast.
                        n_tone = DynamicColor.foreground_tone(bg_tone, n_contrast)
                        f_tone = DynamicColor.foreground_tone(bg_tone, f_contrast)

            if (f_tone - n_tone) * expansion_dir < delta:
                # 2nd round: expand farther to match delta, if contrast is not
                # satisfied.
                f_tone = math_utils.clamp_double(0, 100, n_tone + delta * expansion_dir)
                if (f_tone - n_tone) * expansion_dir >= delta:
                    # Good! Tones now satisfy the constraint; no change needed.
                    pass
                else:
                    # 3rd round: contract nearer to match delta.
                    n_tone = math_utils.clamp_double(0, 100, f_tone - delta * expansion_dir)

            # Avoids the 50-59 awkward zone.
            if 50 <= n_tone < 60:
                # If `nearer` is in the awkward zone, move it away, together with
                # `farther`.
                if expansion_dir > 0:
                    n_tone = 60
                    f_tone = max(f_tone, n_tone + delta * expansion_dir)
                else:
                    n_tone = 49
                    f_tone = min(f_tone, n_tone + delta * expansion_dir)
            elif 50 <= f_tone < 60:
                if stay_together:
                    # Fixes both, to avoid two colors on opposite sides of the "awkward
                    # zone".
                    if expansion_dir > 0:
                        n_tone = 60
                        f_tone = max(f_tone, n_tone + delta * expansion_dir)
                    else:
                        n_tone = 49
                        f_tone = min(f_tone, n_tone + delta * expansion_dir)
                else:
                    # Not required to stay together; fixes just one.
                    if expansion_dir > 0:
                        f_tone = 60
                    else:
                        f_tone = 49

            # Returns `n_tone` if this color is `nearer`, otherwise `f_tone`.
            return n_tone if am_nearer else f_tone
        else:
            # Case 2: No contrast pair; just solve for itself.
            answer = color.tone(scheme)

            if (color.background is None or
                    color.background(scheme) is None or
                    color.contrast_curve is None or
                    color.contrast_curve(scheme) is None):
                return answer  # No adjustment for colors with no background.

            bg_tone = color.background(scheme).get_tone(scheme)
            desired_ratio = color.contrast_curve(scheme).get(scheme.contrast_level)

            if Contrast.ratio_of_tones(bg_tone, answer) >= desired_ratio:
                # Don't "improve" what's good enough.
                pass
            else:
                # Rough improvement.
                answer = DynamicColor.foreground_tone(bg_tone, desired_ratio)

            if decreasing_contrast:
                answer = DynamicColor.foreground_tone(bg_tone, desired_ratio)

            if color.is_background and 50 <= answer < 60:
                # Must adjust
                if Contrast.ratio_of_tones(49, bg_tone) >= desired_ratio:
                    answer = 49
                else:
                    answer = 60

            if (color.second_background is None or
                    color.second_background(scheme) is None):
                return answer

            # Case 3: Adjust for dual backgrounds.
            bg1, bg2 = color.background, color.second_background
            bg_tone1, bg_tone2 = bg1(scheme).get_tone(scheme), bg2(scheme).get_tone(scheme)
            upper, lower = max(bg_tone1, bg_tone2), min(bg_tone1, bg_tone2)

            if (Contrast.ratio_of_tones(upper, answer) >= desired_ratio and
                    Contrast.ratio_of_tones(lower, answer) >= desired_ratio):
                return answer

            # The darkest light tone that satisfies the desired ratio,
            # or -1 if such ratio cannot be reached.
            light_option = Contrast.lighter(upper, desired_ratio)

            # The lightest dark tone that satisfies the desired ratio,
            # or -1 if such ratio cannot be reached.
            dark_option = Contrast.darker(lower, desired_ratio)

            # Tones suitable for the foreground.
            availables = []
            if light_option != -1:
                availables.append(light_option)
            if dark_option != -1:
                availables.append(dark_option)

            prefers_light = (DynamicColor.tone_prefers_light_foreground(bg_tone1) or
                            DynamicColor.tone_prefers_light_foreground(bg_tone2))
            if prefers_light:
                return 100 if light_option < 0 else light_option
            if len(availables) == 1:
                return availables[0]
            return 0 if dark_option < 0 else dark_option


class ColorCalculationDelegateImpl2025(ColorCalculationDelegate):
    """
    A delegate for the color calculation of a DynamicScheme in the 2025 spec.
    """

    def get_hct(self, scheme: "DynamicScheme", color: DynamicColor) -> Hct:
        """
        Get the HCT color for the given scheme and color.
        
        Args:
            scheme: The dynamic scheme.
            color: The dynamic color.
            
        Returns:
            An HCT color.
        """
        palette = color.palette(scheme)
        tone = color.get_tone(scheme)
        hue = palette.hue
        chroma = palette.chroma * (color.chroma_multiplier(scheme) if color.chroma_multiplier else 1)

        return Hct.from_hct(hue, chroma, tone)

    def get_tone(self, scheme: "DynamicScheme", color: DynamicColor) -> float:
        """
        Get the tone for the given scheme and color.
        
        Args:
            scheme: The dynamic scheme.
            color: The dynamic color.
            
        Returns:
            A tone value between 0 and 100.
        """
        tone_delta_pair = color.tone_delta_pair(scheme) if color.tone_delta_pair else None

        # Case 0: tone delta constraint.
        if tone_delta_pair:
            role_a = tone_delta_pair.role_a
            role_b = tone_delta_pair.role_b
            polarity = tone_delta_pair.polarity
            constraint = tone_delta_pair.constraint
            absolute_delta = (-tone_delta_pair.delta if polarity == 'darker' or
                            (polarity == 'relative_lighter' and scheme.is_dark) or
                            (polarity == 'relative_darker' and not scheme.is_dark)
                            else tone_delta_pair.delta)

            am_role_a = color.name == role_a.name
            self_role = role_a if am_role_a else role_b
            ref_role = role_b if am_role_a else role_a
            self_tone = self_role.tone(scheme)
            ref_tone = ref_role.get_tone(scheme)
            relative_delta = absolute_delta * (1 if am_role_a else -1)

            if constraint == 'exact':
                self_tone = math_utils.clamp_double(0, 100, ref_tone + relative_delta)
            elif constraint == 'nearer':
                if relative_delta > 0:
                    self_tone = math_utils.clamp_double(
                        0, 100,
                        math_utils.clamp_double(ref_tone, ref_tone + relative_delta, self_tone))
                else:
                    self_tone = math_utils.clamp_double(
                        0, 100,
                        math_utils.clamp_double(ref_tone + relative_delta, ref_tone, self_tone))
            elif constraint == 'farther':
                if relative_delta > 0:
                    self_tone = math_utils.clamp_double(ref_tone + relative_delta, 100, self_tone)
                else:
                    self_tone = math_utils.clamp_double(0, ref_tone + relative_delta, self_tone)

            if color.background and color.contrast_curve:
                background = color.background(scheme)
                contrast_curve = color.contrast_curve(scheme)
                if background and contrast_curve:
                    # Adjust the tones for contrast, if background and contrast curve
                    # are defined.
                    bg_tone = background.get_tone(scheme)
                    self_contrast = contrast_curve.get(scheme.contrast_level)
                    self_tone = (self_tone if Contrast.ratio_of_tones(bg_tone, self_tone) >= self_contrast and
                                scheme.contrast_level >= 0 else
                                DynamicColor.foreground_tone(bg_tone, self_contrast))

            # This can avoid the awkward tones for background colors including the
            # access fixed colors. Accent fixed dim colors should not be adjusted.
            if color.is_background and not color.name.endswith('_fixed_dim'):
                if self_tone >= 57:
                    self_tone = math_utils.clamp_double(65, 100, self_tone)
                else:
                    self_tone = math_utils.clamp_double(0, 49, self_tone)

            return self_tone
        else:
            # Case 1: No tone delta pair; just solve for itself.
            answer = color.tone(scheme)

            if (color.background is None or
                    color.background(scheme) is None or
                    color.contrast_curve is None or
                    color.contrast_curve(scheme) is None):
                return answer  # No adjustment for colors with no background.

            bg_tone = color.background(scheme).get_tone(scheme)
            desired_ratio = color.contrast_curve(scheme).get(scheme.contrast_level)

            # Recalculate the tone from desired contrast ratio if the current
            # contrast ratio is not enough or desired contrast level is decreasing
            # (<0).
            answer = (answer if Contrast.ratio_of_tones(bg_tone, answer) >= desired_ratio and
                     scheme.contrast_level >= 0 else
                     DynamicColor.foreground_tone(bg_tone, desired_ratio))

            # This can avoid the awkward tones for background colors including the
            # access fixed colors. Accent fixed dim colors should not be adjusted.
            if color.is_background and not color.name.endswith('_fixed_dim'):
                if answer >= 57:
                    answer = math_utils.clamp_double(65, 100, answer)
                else:
                    answer = math_utils.clamp_double(0, 49, answer)

            if (color.second_background is None or
                    color.second_background(scheme) is None):
                return answer

            # Case 2: Adjust for dual backgrounds.
            bg1, bg2 = color.background, color.second_background
            bg_tone1, bg_tone2 = bg1(scheme).get_tone(scheme), bg2(scheme).get_tone(scheme)
            upper, lower = max(bg_tone1, bg_tone2), min(bg_tone1, bg_tone2)

            if (Contrast.ratio_of_tones(upper, answer) >= desired_ratio and
                    Contrast.ratio_of_tones(lower, answer) >= desired_ratio):
                return answer

            # The darkest light tone that satisfies the desired ratio,
            # or -1 if such ratio cannot be reached.
            light_option = Contrast.lighter(upper, desired_ratio)

            # The lightest dark tone that satisfies the desired ratio,
            # or -1 if such ratio cannot be reached.
            dark_option = Contrast.darker(lower, desired_ratio)

            # Tones suitable for the foreground.
            availables = []
            if light_option != -1:
                availables.append(light_option)
            if dark_option != -1:
                availables.append(dark_option)

            prefers_light = (DynamicColor.tone_prefers_light_foreground(bg_tone1) or
                            DynamicColor.tone_prefers_light_foreground(bg_tone2))
            if prefers_light:
                return 100 if light_option < 0 else light_option
            if len(availables) == 1:
                return availables[0]
            return 0 if dark_option < 0 else dark_option


# Global delegates for different spec versions
spec2021 = ColorCalculationDelegateImpl2021()
spec2025 = ColorCalculationDelegateImpl2025()


def get_spec(spec_version: SpecVersion) -> ColorCalculationDelegate:
    """
    Returns the ColorCalculationDelegate for the given spec version.
    
    Args:
        spec_version: The spec version.
        
    Returns:
        A ColorCalculationDelegate.
    """
    return spec2025 if spec_version == '2025' else spec2021


# dynamiccolor/color_spec.py

from typing import Literal

from PyMCUlib.dynamiccolor.color_spec_delegate import ColorSpecDelegate
from PyMCUlib.dynamiccolor.color_spec_2021 import ColorSpecDelegateImpl2021
from PyMCUlib.dynamiccolor.color_spec_2025 import ColorSpecDelegateImpl2025

# Instantiate delegates for each spec version
spec_2021 = ColorSpecDelegateImpl2021()
spec_2025 = ColorSpecDelegateImpl2025()

def get_spec(spec_version: Literal['2021', '2025']) -> ColorSpecDelegate:
    """
    Returns the ColorSpecDelegate for the given spec version.
    
    Args:
        spec_version: The specification version, either '2021' or '2025'
        
    Returns:
        The appropriate ColorSpecDelegate implementation
    """
    if spec_version == '2021':
        return spec_2021
    if spec_version == '2025':
        return spec_2025


# scheme/scheme_fidelity.py
"""
A scheme that places the source color in `Scheme.primary_container`.

Primary Container is the source color, adjusted for color relativity.
It maintains constant appearance in light mode and dark mode.
This adds ~5 tone in light mode, and subtracts ~5 tone in dark mode.
Tertiary Container is the complement to the source color, using
`TemperatureCache`. It also maintains constant appearance.
"""

from PyMCUlib.dynamiccolor.dynamic_scheme import DynamicScheme
from PyMCUlib.dynamiccolor.variant import Variant
from PyMCUlib.hct.hct import Hct


class SchemeFidelity(DynamicScheme):
    """
    A scheme that places the source color in `Scheme.primary_container`.

    Primary Container is the source color, adjusted for color relativity.
    It maintains constant appearance in light mode and dark mode.
    This adds ~5 tone in light mode, and subtracts ~5 tone in dark mode.
    Tertiary Container is the complement to the source color, using
    `TemperatureCache`. It also maintains constant appearance.
    """

    DEFAULT_SPEC_VERSION = '2021'
    DEFAULT_PLATFORM = 'phone'

    def __init__(self, source_color_hct: Hct, is_dark: bool, contrast_level: float):
        """
        Construct a new SchemeFidelity instance.

        Args:
            source_color_hct: The source color of the theme as an HCT color.
            is_dark: Whether the scheme is in dark mode or light mode.
            contrast_level: The contrast level of the theme.
        """
        super().__init__({
            'source_color_hct': source_color_hct,
            'variant': Variant.FIDELITY,
            'contrast_level': contrast_level,
            'is_dark': is_dark,
            'platform': SchemeFidelity.DEFAULT_PLATFORM,
            'spec_version': SchemeFidelity.DEFAULT_SPEC_VERSION,
        })


# utils/math_utils.py

from typing import List


def signum(num: float) -> int:
    """
    The signum function.

    Args:
        num: The input number

    Returns:
        1 if num > 0, -1 if num < 0, and 0 if num = 0
    """
    if num < 0:
        return -1
    elif num == 0:
        return 0
    else:
        return 1


def lerp(start: float, stop: float, amount: float) -> float:
    """
    The linear interpolation function.

    Args:
        start: The start value
        stop: The end value
        amount: The interpolation factor (0.0 to 1.0)

    Returns:
        start if amount = 0 and stop if amount = 1
    """
    return (1.0 - amount) * start + amount * stop


def clamp_int(min_val: int, max_val: int, input_val: int) -> int:
    """
    Clamps an integer between two integers.

    Args:
        min_val: The lower bound
        max_val: The upper bound
        input_val: The input value

    Returns:
        input_val when min_val <= input_val <= max_val, and either min_val or max_val
        otherwise.
    """
    if input_val < min_val:
        return min_val
    elif input_val > max_val:
        return max_val

    return input_val


def clamp_double(min_val: float, max_val: float, input_val: float) -> float:
    """
    Clamps a floating-point number between two floating-point numbers.

    Args:
        min_val: The lower bound
        max_val: The upper bound
        input_val: The input value

    Returns:
        input_val when min_val <= input_val <= max_val, and either min_val or max_val
        otherwise.
    """
    if input_val < min_val:
        return min_val
    elif input_val > max_val:
        return max_val

    return input_val


def sanitize_degrees_int(degrees: int) -> int:
    """
    Sanitizes a degree measure as an integer.

    Args:
        degrees: The input angle in degrees

    Returns:
        A degree measure between 0 (inclusive) and 360 (exclusive).
    """
    degrees = degrees % 360
    if degrees < 0:
        degrees = degrees + 360
    return degrees


def sanitize_degrees_double(degrees: float) -> float:
    """
    Sanitizes a degree measure as a floating-point number.

    Args:
        degrees: The input angle in degrees

    Returns:
        A degree measure between 0.0 (inclusive) and 360.0 (exclusive).
    """
    degrees = degrees % 360.0
    if degrees < 0:
        degrees = degrees + 360.0
    return degrees


def rotation_direction(from_val: float, to_val: float) -> float:
    """
    Sign of direction change needed to travel from one angle to another.

    For angles that are 180 degrees apart from each other, both
    directions have the same travel distance, so either direction is
    shortest. The value 1.0 is returned in this case.

    Args:
        from_val: The angle travel starts from, in degrees.
        to_val: The angle travel ends at, in degrees.

    Returns:
        -1 if decreasing from_val leads to the shortest travel
        distance, 1 if increasing from_val leads to the shortest travel
        distance.
    """
    increasing_difference = sanitize_degrees_double(to_val - from_val)
    return 1.0 if increasing_difference <= 180.0 else -1.0


def difference_degrees(a: float, b: float) -> float:
    """
    Distance of two points on a circle, represented using degrees.

    Args:
        a: First angle in degrees
        b: Second angle in degrees

    Returns:
        The shortest angular distance between the two angles
    """
    return 180.0 - abs(abs(a - b) - 180.0)


def matrix_multiply(row: List[float], matrix: List[List[float]]) -> List[float]:
    """
    Multiplies a 1x3 row vector with a 3x3 matrix.

    Args:
        row: A 1x3 row vector
        matrix: A 3x3 matrix

    Returns:
        The resulting 1x3 row vector after multiplication
    """
    a = row[0] * matrix[0][0] + row[1] * matrix[0][1] + row[2] * matrix[0][2]
    b = row[0] * matrix[1][0] + row[1] * matrix[1][1] + row[2] * matrix[1][2]
    c = row[0] * matrix[2][0] + row[1] * matrix[2][1] + row[2] * matrix[2][2]
    return [a, b, c]


# scheme/scheme.py

from typing import Dict
from PyMCUlib.palettes.core_palette import CorePalette


class Scheme:
    """
    DEPRECATED. The `Scheme` class is deprecated in favor of `DynamicScheme`.

    Represents a Material color scheme, a mapping of color roles to colors.
    """

    def __init__(self, props: Dict[str, int]) -> None:
        """
        Private constructor for Scheme class.

        Args:
            props: A dictionary containing all color properties for the scheme.
        """
        self._props = props

    @property
    def primary(self) -> int:
        """Primary color."""
        return self._props["primary"]

    @property
    def on_primary(self) -> int:
        """Color used for text and icons on primary color."""
        return self._props["on_primary"]

    @property
    def primary_container(self) -> int:
        """Container color for the primary color."""
        return self._props["primary_container"]

    @property
    def on_primary_container(self) -> int:
        """Color used for text and icons on primary_container color."""
        return self._props["on_primary_container"]

    @property
    def secondary(self) -> int:
        """Secondary color."""
        return self._props["secondary"]

    @property
    def on_secondary(self) -> int:
        """Color used for text and icons on secondary color."""
        return self._props["on_secondary"]

    @property
    def secondary_container(self) -> int:
        """Container color for the secondary color."""
        return self._props["secondary_container"]

    @property
    def on_secondary_container(self) -> int:
        """Color used for text and icons on secondary_container color."""
        return self._props["on_secondary_container"]

    @property
    def tertiary(self) -> int:
        """Tertiary color."""
        return self._props["tertiary"]

    @property
    def on_tertiary(self) -> int:
        """Color used for text and icons on tertiary color."""
        return self._props["on_tertiary"]

    @property
    def tertiary_container(self) -> int:
        """Container color for the tertiary color."""
        return self._props["tertiary_container"]

    @property
    def on_tertiary_container(self) -> int:
        """Color used for text and icons on tertiary_container color."""
        return self._props["on_tertiary_container"]

    @property
    def error(self) -> int:
        """Error color."""
        return self._props["error"]

    @property
    def on_error(self) -> int:
        """Color used for text and icons on error color."""
        return self._props["on_error"]

    @property
    def error_container(self) -> int:
        """Container color for the error color."""
        return self._props["error_container"]

    @property
    def on_error_container(self) -> int:
        """Color used for text and icons on error_container color."""
        return self._props["on_error_container"]

    @property
    def background(self) -> int:
        """Background color."""
        return self._props["background"]

    @property
    def on_background(self) -> int:
        """Color used for text and icons on background color."""
        return self._props["on_background"]

    @property
    def surface(self) -> int:
        """Surface color."""
        return self._props["surface"]

    @property
    def on_surface(self) -> int:
        """Color used for text and icons on surface color."""
        return self._props["on_surface"]

    @property
    def surface_variant(self) -> int:
        """Surface variant color."""
        return self._props["surface_variant"]

    @property
    def on_surface_variant(self) -> int:
        """Color used for text and icons on surface_variant color."""
        return self._props["on_surface_variant"]

    @property
    def outline(self) -> int:
        """Outline color."""
        return self._props["outline"]

    @property
    def outline_variant(self) -> int:
        """Outline variant color."""
        return self._props["outline_variant"]

    @property
    def shadow(self) -> int:
        """Shadow color."""
        return self._props["shadow"]

    @property
    def scrim(self) -> int:
        """Scrim color."""
        return self._props["scrim"]

    @property
    def inverse_surface(self) -> int:
        """Inverse surface color."""
        return self._props["inverse_surface"]

    @property
    def inverse_on_surface(self) -> int:
        """Color used for text and icons on inverse_surface color."""
        return self._props["inverse_on_surface"]

    @property
    def inverse_primary(self) -> int:
        """Inverse primary color."""
        return self._props["inverse_primary"]

    @staticmethod
    def light(argb: int) -> 'Scheme':
        """
        Creates a light Material color scheme, based on the color's hue.

        Args:
            argb: ARGB representation of a color.

        Returns:
            Light Material color scheme, based on the color's hue.
        """
        return Scheme.light_from_core_palette(CorePalette.of(argb))

    @staticmethod
    def dark(argb: int) -> 'Scheme':
        """
        Creates a dark Material color scheme, based on the color's hue.

        Args:
            argb: ARGB representation of a color.

        Returns:
            Dark Material color scheme, based on the color's hue.
        """
        return Scheme.dark_from_core_palette(CorePalette.of(argb))

    @staticmethod
    def light_content(argb: int) -> 'Scheme':
        """
        Creates a light Material content color scheme, based on the color's hue.

        Args:
            argb: ARGB representation of a color.

        Returns:
            Light Material content color scheme, based on the color's hue.
        """
        return Scheme.light_from_core_palette(CorePalette.content_of(argb))

    @staticmethod
    def dark_content(argb: int) -> 'Scheme':
        """
        Creates a dark Material content color scheme, based on the color's hue.

        Args:
            argb: ARGB representation of a color.

        Returns:
            Dark Material content color scheme, based on the color's hue.
        """
        return Scheme.dark_from_core_palette(CorePalette.content_of(argb))

    @staticmethod
    def light_from_core_palette(core: CorePalette) -> 'Scheme':
        """
        Creates a light scheme from a core palette.

        Args:
            core: A CorePalette.

        Returns:
            A light scheme.
        """
        return Scheme({
            "primary": core.a1.tone(40),
            "on_primary": core.a1.tone(100),
            "primary_container": core.a1.tone(90),
            "on_primary_container": core.a1.tone(10),
            "secondary": core.a2.tone(40),
            "on_secondary": core.a2.tone(100),
            "secondary_container": core.a2.tone(90),
            "on_secondary_container": core.a2.tone(10),
            "tertiary": core.a3.tone(40),
            "on_tertiary": core.a3.tone(100),
            "tertiary_container": core.a3.tone(90),
            "on_tertiary_container": core.a3.tone(10),
            "error": core.error.tone(40),
            "on_error": core.error.tone(100),
            "error_container": core.error.tone(90),
            "on_error_container": core.error.tone(10),
            "background": core.n1.tone(99),
            "on_background": core.n1.tone(10),
            "surface": core.n1.tone(99),
            "on_surface": core.n1.tone(10),
            "surface_variant": core.n2.tone(90),
            "on_surface_variant": core.n2.tone(30),
            "outline": core.n2.tone(50),
            "outline_variant": core.n2.tone(80),
            "shadow": core.n1.tone(0),
            "scrim": core.n1.tone(0),
            "inverse_surface": core.n1.tone(20),
            "inverse_on_surface": core.n1.tone(95),
            "inverse_primary": core.a1.tone(80)
        })

    @staticmethod
    def dark_from_core_palette(core: CorePalette) -> 'Scheme':
        """
        Creates a dark scheme from a core palette.

        Args:
            core: A CorePalette.

        Returns:
            A dark scheme.
        """
        return Scheme({
            "primary": core.a1.tone(80),
            "on_primary": core.a1.tone(20),
            "primary_container": core.a1.tone(30),
            "on_primary_container": core.a1.tone(90),
            "secondary": core.a2.tone(80),
            "on_secondary": core.a2.tone(20),
            "secondary_container": core.a2.tone(30),
            "on_secondary_container": core.a2.tone(90),
            "tertiary": core.a3.tone(80),
            "on_tertiary": core.a3.tone(20),
            "tertiary_container": core.a3.tone(30),
            "on_tertiary_container": core.a3.tone(90),
            "error": core.error.tone(80),
            "on_error": core.error.tone(20),
            "error_container": core.error.tone(30),
            "on_error_container": core.error.tone(80),
            "background": core.n1.tone(10),
            "on_background": core.n1.tone(90),
            "surface": core.n1.tone(10),
            "on_surface": core.n1.tone(90),
            "surface_variant": core.n2.tone(30),
            "on_surface_variant": core.n2.tone(80),
            "outline": core.n2.tone(60),
            "outline_variant": core.n2.tone(30),
            "shadow": core.n1.tone(0),
            "scrim": core.n1.tone(0),
            "inverse_surface": core.n1.tone(90),
            "inverse_on_surface": core.n1.tone(20),
            "inverse_primary": core.a1.tone(40)
        })

    def to_json(self) -> Dict[str, int]:
        """
        Convert the scheme to a JSON serializable dictionary.

        Returns:
            A dictionary representation of the scheme.
        """
        return {**self._props}


# utils/color_utils.py

from typing import List
import math
from PyMCUlib.utils import math_utils

# Constants for color space conversions
SRGB_TO_XYZ = [
    [0.41233895, 0.35762064, 0.18051042],
    [0.2126, 0.7152, 0.0722],
    [0.01932141, 0.11916382, 0.95034478],
]

XYZ_TO_SRGB = [
    [
        3.2413774792388685,
        -1.5376652402851851,
        -0.49885366846268053,
    ],
    [
        -0.9691452513005321,
        1.8758853451067872,
        0.04156585616912061,
    ],
    [
        0.05562093689691305,
        -0.20395524564742123,
        1.0571799111220335,
    ],
]

WHITE_POINT_D65 = [95.047, 100.0, 108.883]


def argb_from_rgb(red: int, green: int, blue: int) -> int:
    """
    Converts a color from RGB components to ARGB format.

    Args:
        red: The red component (0-255)
        green: The green component (0-255)
        blue: The blue component (0-255)

    Returns:
        The ARGB representation of the color
    """
    return (255 << 24 | (red & 255) << 16 | (green & 255) << 8 | blue & 255) & 0xFFFFFFFF


def argb_from_linrgb(linrgb: List[float]) -> int:
    """
    Converts a color from linear RGB components to ARGB format.

    Args:
        linrgb: The linear RGB components

    Returns:
        The ARGB representation of the color
    """
    r = delinearized(linrgb[0])
    g = delinearized(linrgb[1])
    b = delinearized(linrgb[2])
    return argb_from_rgb(r, g, b)


def alpha_from_argb(argb: int) -> int:
    """
    Returns the alpha component of a color in ARGB format.

    Args:
        argb: The ARGB color

    Returns:
        The alpha component (0-255)
    """
    return (argb >> 24) & 255


def red_from_argb(argb: int) -> int:
    """
    Returns the red component of a color in ARGB format.

    Args:
        argb: The ARGB color

    Returns:
        The red component (0-255)
    """
    return (argb >> 16) & 255


def green_from_argb(argb: int) -> int:
    """
    Returns the green component of a color in ARGB format.

    Args:
        argb: The ARGB color

    Returns:
        The green component (0-255)
    """
    return (argb >> 8) & 255


def blue_from_argb(argb: int) -> int:
    """
    Returns the blue component of a color in ARGB format.

    Args:
        argb: The ARGB color

    Returns:
        The blue component (0-255)
    """
    return argb & 255


def is_opaque(argb: int) -> bool:
    """
    Returns whether a color in ARGB format is opaque.

    Args:
        argb: The ARGB color

    Returns:
        True if the color is opaque, False otherwise
    """
    return alpha_from_argb(argb) >= 255


def argb_from_xyz(x: float, y: float, z: float) -> int:
    """
    Converts a color from XYZ to ARGB.

    Args:
        x: The X component
        y: The Y component
        z: The Z component

    Returns:
        The ARGB representation of the color
    """
    matrix = XYZ_TO_SRGB
    linear_r = matrix[0][0] * x + matrix[0][1] * y + matrix[0][2] * z
    linear_g = matrix[1][0] * x + matrix[1][1] * y + matrix[1][2] * z
    linear_b = matrix[2][0] * x + matrix[2][1] * y + matrix[2][2] * z
    r = delinearized(linear_r)
    g = delinearized(linear_g)
    b = delinearized(linear_b)
    return argb_from_rgb(r, g, b)


def xyz_from_argb(argb: int) -> List[float]:
    """
    Converts a color from ARGB to XYZ.

    Args:
        argb: The ARGB color

    Returns:
        The XYZ components as a list [x, y, z]
    """
    r = linearized(red_from_argb(argb))
    g = linearized(green_from_argb(argb))
    b = linearized(blue_from_argb(argb))
    return math_utils.matrix_multiply([r, g, b], SRGB_TO_XYZ)


def argb_from_lab(l: float, a: float, b_val: float) -> int:
    """
    Converts a color represented in Lab color space into an ARGB
    integer.

    Args:
        l: The L component
        a: The a component
        b_val: The b component

    Returns:
        The ARGB representation of the color
    """
    white_point = WHITE_POINT_D65
    fy = (l + 16.0) / 116.0
    fx = a / 500.0 + fy
    fz = fy - b_val / 200.0
    x_normalized = _lab_invf(fx)
    y_normalized = _lab_invf(fy)
    z_normalized = _lab_invf(fz)
    x = x_normalized * white_point[0]
    y = y_normalized * white_point[1]
    z = z_normalized * white_point[2]
    return argb_from_xyz(x, y, z)


def lab_from_argb(argb: int) -> List[float]:
    """
    Converts a color from ARGB representation to L*a*b*
    representation.

    Args:
        argb: The ARGB representation of a color

    Returns:
        A Lab object representing the color as [L, a, b]
    """
    linear_r = linearized(red_from_argb(argb))
    linear_g = linearized(green_from_argb(argb))
    linear_b = linearized(blue_from_argb(argb))
    matrix = SRGB_TO_XYZ
    x = matrix[0][0] * linear_r + matrix[0][1] * linear_g + matrix[0][2] * linear_b
    y = matrix[1][0] * linear_r + matrix[1][1] * linear_g + matrix[1][2] * linear_b
    z = matrix[2][0] * linear_r + matrix[2][1] * linear_g + matrix[2][2] * linear_b
    white_point = WHITE_POINT_D65
    x_normalized = x / white_point[0]
    y_normalized = y / white_point[1]
    z_normalized = z / white_point[2]
    fx = _lab_f(x_normalized)
    fy = _lab_f(y_normalized)
    fz = _lab_f(z_normalized)
    l = 116.0 * fy - 16.0
    a = 500.0 * (fx - fy)
    b = 200.0 * (fy - fz)
    return [l, a, b]


def argb_from_lstar(lstar: float) -> int:
    """
    Converts an L* value to an ARGB representation.

    Args:
        lstar: L* in L*a*b*

    Returns:
        ARGB representation of grayscale color with lightness
        matching L*
    """
    y = y_from_lstar(lstar)
    component = delinearized(y)
    return argb_from_rgb(component, component, component)


def lstar_from_argb(argb: int) -> float:
    """
    Computes the L* value of a color in ARGB representation.

    Args:
        argb: ARGB representation of a color

    Returns:
        L*, from L*a*b*, coordinate of the color
    """
    y = xyz_from_argb(argb)[1]
    return 116.0 * _lab_f(y / 100.0) - 16.0


def y_from_lstar(lstar: float) -> float:
    """
    Converts an L* value to a Y value.

    L* in L*a*b* and Y in XYZ measure the same quantity, luminance.

    L* measures perceptual luminance, a linear scale. Y in XYZ
    measures relative luminance, a logarithmic scale.

    Args:
        lstar: L* in L*a*b*

    Returns:
        Y in XYZ
    """
    return 100.0 * _lab_invf((lstar + 16.0) / 116.0)


def lstar_from_y(y: float) -> float:
    """
    Converts a Y value to an L* value.

    L* in L*a*b* and Y in XYZ measure the same quantity, luminance.

    L* measures perceptual luminance, a linear scale. Y in XYZ
    measures relative luminance, a logarithmic scale.

    Args:
        y: Y in XYZ

    Returns:
        L* in L*a*b*
    """
    return _lab_f(y / 100.0) * 116.0 - 16.0


def linearized(rgb_component: int) -> float:
    """
    Linearizes an RGB component.

    Args:
        rgb_component: 0 <= rgb_component <= 255, represents R/G/B channel

    Returns:
        0.0 <= output <= 100.0, color channel converted to linear RGB space
    """
    normalized = rgb_component / 255.0
    if normalized <= 0.040449936:
        return normalized / 12.92 * 100.0
    else:
        return math.pow((normalized + 0.055) / 1.055, 2.4) * 100.0


def delinearized(rgb_component: float) -> int:
    """
    Delinearizes an RGB component.

    Args:
        rgb_component: 0.0 <= rgb_component <= 100.0, represents linear R/G/B channel

    Returns:
        0 <= output <= 255, color channel converted to regular RGB space
    """
    normalized = rgb_component / 100.0
    delinearized_val = 0.0
    if normalized <= 0.0031308:
        delinearized_val = normalized * 12.92
    else:
        delinearized_val = 1.055 * math.pow(normalized, 1.0 / 2.4) - 0.055
    return math_utils.clamp_int(0, 255, round(delinearized_val * 255.0))


def white_point_d65() -> List[float]:
    """
    Returns the standard white point; white on a sunny day.

    Returns:
        The white point
    """
    return WHITE_POINT_D65


def _lab_f(t: float) -> float:
    """
    Helper function for Lab color space conversions.

    Args:
        t: Input value

    Returns:
        Transformed value
    """
    e = 216.0 / 24389.0
    kappa = 24389.0 / 27.0
    if t > e:
        return math.pow(t, 1.0 / 3.0)
    else:
        return (kappa * t + 16.0) / 116.0


def _lab_invf(ft: float) -> float:
    """
    Inverse of _lab_f function.

    Args:
        ft: Input value

    Returns:
        Inverse transformed value
    """
    e = 216.0 / 24389.0
    kappa = 24389.0 / 27.0
    ft3 = ft * ft * ft
    if ft3 > e:
        return ft3
    else:
        return (116.0 * ft - 16.0) / kappa


# utils/theme_utils.py

from typing import List, Dict, Any, TypedDict, Optional, Union
import re
from PIL import Image
from PyMCUlib.palettes.core_palette import CorePalette
from PyMCUlib.scheme.scheme import Scheme
from PyMCUlib.palettes.tonal_palette import TonalPalette
from PyMCUlib.blend.blend import Blend
from PyMCUlib.utils import image_utils
from PyMCUlib.utils import string_utils


class CustomColor(TypedDict):
    """Custom color used to pair with a theme"""
    value: int
    name: str
    blend: bool


class ColorGroup(TypedDict):
    """Color group containing main colors and container colors"""
    color: int
    onColor: int
    colorContainer: int
    onColorContainer: int


class CustomColorGroup(TypedDict):
    """Custom color group containing light and dark variants"""
    color: CustomColor
    value: int
    light: ColorGroup
    dark: ColorGroup


class Theme(TypedDict):
    """Theme containing color schemes and palettes"""
    source: int
    schemes: Dict[str, Any]  # Contains light and dark schemes
    palettes: Dict[str, TonalPalette]  # Contains various palettes
    customColors: List[CustomColorGroup]

class ApplyThemeOptions(TypedDict, total=False):
    """Options for apply_theme"""
    dark: bool
    target: Any
    brightnessSuffix: bool
    paletteTones: List[int]


def theme_from_source_color(
        source: int, custom_colors: List[CustomColor] = None) -> Theme:
    """
    Generate a theme from a source color

    Args:
        source: Source color in ARGB format
        custom_colors: Array of custom colors (optional)

    Returns:
        Theme object
    """
    if custom_colors is None:
        custom_colors = []

    palette = CorePalette.of(source)
    return {
        "source": source,
        "schemes": {
            "light": Scheme.light(source),
            "dark": Scheme.dark(source),
        },
        "palettes": {
            "primary": palette.a1,
            "secondary": palette.a2,
            "tertiary": palette.a3,
            "neutral": palette.n1,
            "neutralVariant": palette.n2,
            "error": palette.error,
        },
        "customColors": [custom_color(source, c) for c in custom_colors],
    }


def theme_from_image(
    image: Union[str, bytes, Image.Image],
    custom_colors: Optional[List[CustomColor]] = None) -> Theme:
    """
    Generate a theme from an image source

    Args:
        image: Image data (file path, bytes, or PIL Image)
        custom_colors: Array of custom colors (optional)

    Returns:
        Theme object
    """
    if custom_colors is None:
        custom_colors = []

    source = image_utils.source_color_from_image(image)
    return theme_from_source_color(source, custom_colors)


def custom_color(source: int, color: CustomColor) -> CustomColorGroup:
    """
    Generate custom color group from source and target color

    Args:
        source: Source color in ARGB format
        color: Custom color definition

    Returns:
        Custom color group

    Link: https://m3.material.io/styles/color/the-color-system/color-roles
    """
    value = color["value"]
    from_color = value
    to_color = source
    if color["blend"]:
        value = Blend.harmonize(from_color, to_color)

    palette = CorePalette.of(value)
    tones = palette.a1
    
    return {
        "color": color,
        "value": value,
        "light": {
            "color": tones.tone(40),
            "onColor": tones.tone(100),
            "colorContainer": tones.tone(90),
            "onColorContainer": tones.tone(10),
        },
        "dark": {
            "color": tones.tone(80),
            "onColor": tones.tone(20),
            "colorContainer": tones.tone(30),
            "onColorContainer": tones.tone(90),
        },
    }


def apply_theme(
    theme: Theme,
    options: Optional[ApplyThemeOptions] = None) -> Dict[str, str]:
    """
    Apply a theme to a target.
    
    This function is platform-specific and needs adaptation depending on
    the environment (web, mobile, etc.).

    Args:
        theme: Theme object
        options: Options for applying the theme
            - dark: Whether to use dark scheme (default: False)
            - target: Target to apply the theme to (platform-specific)
            - brightnessSuffix: Whether to add brightness suffixes (default: False)
            - paletteTones: List of tone values to include (optional)

    Returns:
        Dictionary of CSS variables
    """
    if options is None:
        options = {}
    
    # This is a placeholder implementation that returns CSS variables
    # The actual implementation will depend on the target platform
    
    is_dark = options.get("dark", False)
    scheme = theme["schemes"]["dark"] if is_dark else theme["schemes"]["light"]
    
    css_vars = {}
    
    # Add scheme properties
    for key, value in scheme.to_json().items():
        token = re.sub(r'([a-z])([A-Z])', r'\1-\2', key).lower()
        color = string_utils.hex_from_argb(value)
        css_vars[f"--md-sys-color-{token}"] = color
    
    # Add brightness suffix if specified
    if options.get("brightnessSuffix", False):
        for key, value in theme["schemes"]["dark"].to_json().items():
            token = re.sub(r'([a-z])([A-Z])', r'\1-\2', key).lower()
            color = string_utils.hex_from_argb(value)
            css_vars[f"--md-sys-color-{token}-dark"] = color
            
        for key, value in theme["schemes"]["light"].to_json().items():
            token = re.sub(r'([a-z])([A-Z])', r'\1-\2', key).lower()
            color = string_utils.hex_from_argb(value)
            css_vars[f"--md-sys-color-{token}-light"] = color
    
    # Add palette tones if specified
    palette_tones = options.get("paletteTones", [])
    if palette_tones:
        for key, palette in theme["palettes"].items():
            palette_key = re.sub(r'([a-z])([A-Z])', r'\1-\2', key).lower()
            for tone in palette_tones:
                token = f"--md-ref-palette-{palette_key}-{palette_key}{tone}"
                color = string_utils.hex_from_argb(palette.tone(tone))
                css_vars[token] = color
    
    return css_vars


def set_scheme_properties(
        css_vars: Dict[str, str],
        scheme: Any,
        suffix: str = "") -> Dict[str, str]:
    """
    Helper function to set CSS properties for a color scheme

    Args:
        css_vars: Dictionary of CSS variables to update
        scheme: Color scheme
        suffix: Optional suffix to append to CSS variable names

    Returns:
        Updated CSS variables dictionary
    """
    for key, value in scheme.to_json().items():
        token = re.sub(r'([a-z])([A-Z])', r'\1-\2', key).lower()
        color = string_utils.hex_from_argb(value)
        css_vars[f"--md-sys-color-{token}{suffix}"] = color
    
    return css_vars


# hct/__init__.py

# --- viewing_conditions ---
from PyMCUlib.hct.viewing_conditions import ViewingConditions

# --- cam16 ---
from PyMCUlib.hct.cam16 import Cam16

# --- hct_solver ---
from PyMCUlib.hct.hct_solver import HctSolver

# --- hct ---
from PyMCUlib.hct.hct import Hct

__all__ = [
    "ViewingConditions",
    "Cam16",
    "HctSolver",
    "Hct",
]


# quantize/quantizer_wsmeans.py

from typing import Dict, List
import random
from PyMCUlib.quantize.lab_point_provider import LabPointProvider

# Constants
MAX_ITERATIONS = 10
MIN_MOVEMENT_DISTANCE = 3.0


class DistanceAndIndex:
    """
    A wrapper for maintaining a table of distances between K-Means clusters.
    """
    def __init__(self):
        self.distance = -1
        self.index = -1
        
    def __lt__(self, other):
        """
        Allows for sorting DistanceAndIndex objects by distance.
        """
        return self.distance < other.distance


class QuantizerWsmeans:
    """
    An image quantizer that improves on the speed of a standard K-Means algorithm
    by implementing several optimizations, including deduping identical pixels
    and a triangle inequality rule that reduces the number of comparisons needed
    to identify which cluster a point should be moved to.

    Wsmeans stands for Weighted Square Means.

    This algorithm was designed by M. Emre Celebi, and was found in their 2011
    paper, Improving the Performance of K-Means for Color Quantization.
    https://arxiv.org/abs/1101.0395
    """

    @staticmethod
    def quantize(
        input_pixels: List[int], 
        starting_clusters: List[int],
        max_colors: int
    ) -> Dict[int, int]:
        """
        Args:
            input_pixels: Colors in ARGB format.
            starting_clusters: Defines the initial state of the quantizer. Passing
                an empty array is fine, the implementation will create its own initial
                state that leads to reproducible results for the same inputs.
                Passing an array that is the result of Wu quantization leads to higher
                quality results.
            max_colors: The number of colors to divide the image into. A lower
                number of colors may be returned.
            
        Returns:
            Map with keys of colors in ARGB format, and values of number of
            pixels in the original image that correspond to the color in the
            quantized image.
        """
        pixel_to_count = {}
        points = []
        pixels = []
        point_provider = LabPointProvider()
        point_count = 0
        
        # Deduplicate pixels
        for input_pixel in input_pixels:
            pixel_count = pixel_to_count.get(input_pixel)
            if pixel_count is None:
                point_count += 1
                points.append(point_provider.from_int(input_pixel))
                pixels.append(input_pixel)
                pixel_to_count[input_pixel] = 1
            else:
                pixel_to_count[input_pixel] = pixel_count + 1
        
        counts = [0] * point_count
        for i in range(point_count):
            pixel = pixels[i]
            count = pixel_to_count.get(pixel)
            if count is not None:
                counts[i] = count
        
        cluster_count = min(max_colors, point_count)
        if len(starting_clusters) > 0:
            cluster_count = min(cluster_count, len(starting_clusters))
        
        # Initialize clusters
        clusters = []
        for i in range(len(starting_clusters)):
            clusters.append(point_provider.from_int(starting_clusters[i]))
        
        additional_clusters_needed = cluster_count - len(clusters)
        if len(starting_clusters) == 0 and additional_clusters_needed > 0:
            for i in range(additional_clusters_needed):
                l = random.random() * 100.0
                a = random.random() * (100.0 - (-100.0) + 1) + -100
                b = random.random() * (100.0 - (-100.0) + 1) + -100
                clusters.append([l, a, b])
        
        cluster_indices = []
        for i in range(point_count):
            cluster_indices.append(random.randint(0, cluster_count - 1))
        
        # Initialize index matrix and distance-to-index matrix
        index_matrix = []
        for i in range(cluster_count):
            index_matrix.append([0] * cluster_count)
        
        distance_to_index_matrix = []
        for i in range(cluster_count):
            distance_to_index_matrix.append([DistanceAndIndex() for _ in range(cluster_count)])
        
        pixel_count_sums = [0] * cluster_count
        
        # Main iteration loop
        for iteration in range(MAX_ITERATIONS):
            # Calculate distances between clusters
            for i in range(cluster_count):
                for j in range(i + 1, cluster_count):
                    distance = point_provider.distance(clusters[i], clusters[j])
                    distance_to_index_matrix[j][i].distance = distance
                    distance_to_index_matrix[j][i].index = i
                    distance_to_index_matrix[i][j].distance = distance
                    distance_to_index_matrix[i][j].index = j
                
                # Sort distances
                distance_to_index_matrix[i].sort()
                
                # Update index matrix
                for j in range(cluster_count):
                    index_matrix[i][j] = distance_to_index_matrix[i][j].index
            
            points_moved = 0
            # Assign points to clusters
            for i in range(point_count):
                point = points[i]
                previous_cluster_index = cluster_indices[i]
                previous_cluster = clusters[previous_cluster_index]
                previous_distance = point_provider.distance(point, previous_cluster)
                minimum_distance = previous_distance
                new_cluster_index = -1
                
                # Use triangle inequality to reduce comparisons
                for j in range(cluster_count):
                    if (distance_to_index_matrix[previous_cluster_index][j].distance >= 
                            4 * previous_distance):
                        continue
                    
                    distance = point_provider.distance(point, clusters[j])
                    if distance < minimum_distance:
                        minimum_distance = distance
                        new_cluster_index = j
                
                if new_cluster_index != -1:
                    distance_change = abs(
                        (minimum_distance ** 0.5) - (previous_distance ** 0.5)
                    )
                    if distance_change > MIN_MOVEMENT_DISTANCE:
                        points_moved += 1
                        cluster_indices[i] = new_cluster_index
            
            # Check for convergence
            if points_moved == 0 and iteration != 0:
                break
            
            # Update clusters
            component_a_sums = [0] * cluster_count
            component_b_sums = [0] * cluster_count
            component_c_sums = [0] * cluster_count
            
            for i in range(cluster_count):
                pixel_count_sums[i] = 0
            
            for i in range(point_count):
                cluster_index = cluster_indices[i]
                point = points[i]
                count = counts[i]
                pixel_count_sums[cluster_index] += count
                component_a_sums[cluster_index] += (point[0] * count)
                component_b_sums[cluster_index] += (point[1] * count)
                component_c_sums[cluster_index] += (point[2] * count)
            
            for i in range(cluster_count):
                count = pixel_count_sums[i]
                if count == 0:
                    clusters[i] = [0.0, 0.0, 0.0]
                    continue
                
                a = component_a_sums[i] / count
                b = component_b_sums[i] / count
                c = component_c_sums[i] / count
                clusters[i] = [a, b, c]
        
        # Generate result mapping
        argb_to_population = {}
        for i in range(cluster_count):
            count = pixel_count_sums[i]
            if count == 0:
                continue
            
            possible_new_cluster = point_provider.to_int(clusters[i])
            if possible_new_cluster in argb_to_population:
                continue
            
            argb_to_population[possible_new_cluster] = count
        
        return argb_to_population


# dynamiccolor/variant.py

"""
Set of themes supported by Dynamic Color.
Instantiate the corresponding subclass, ex. SchemeTonalSpot, to create
colors corresponding to the theme.
"""

from enum import IntEnum

class Variant(IntEnum):
    """
    Set of themes supported by Dynamic Color.
    Instantiate the corresponding subclass, ex. SchemeTonalSpot, to create
    colors corresponding to the theme.
    """
    MONOCHROME = 0
    NEUTRAL = 1
    TONAL_SPOT = 2
    VIBRANT = 3
    EXPRESSIVE = 4
    FIDELITY = 5
    CONTENT = 6
    RAINBOW = 7
    FRUIT_SALAD = 8


# quantize/quantizer_map.py

from typing import Dict, List
from PyMCUlib.utils import color_utils

class QuantizerMap:
    """
    Quantizes an image into a map, with keys of ARGB colors, and values of the
    number of times that color appears in the image.
    """
    
    @staticmethod
    def quantize(pixels: List[int]) -> Dict[int, int]:
        """
        Args:
            pixels: Colors in ARGB format.
        
        Returns:
            A Dict with keys of ARGB colors, and values of the number of times
            the color appears in the image.
        """
        count_by_color = {}
        for pixel in pixels:
            alpha = color_utils.alpha_from_argb(pixel)
            if alpha < 255:
                continue
            
            count_by_color[pixel] = count_by_color.get(pixel, 0) + 1
                
        return count_by_color


# scheme/scheme_expressive.py
"""
A Dynamic Color theme that is intentionally detached from the source color.
"""

from typing import Optional

from PyMCUlib.dynamiccolor.color_spec_delegate import SpecVersion
from PyMCUlib.dynamiccolor.dynamic_scheme import DynamicScheme, Platform
from PyMCUlib.dynamiccolor.variant import Variant
from PyMCUlib.hct.hct import Hct


class SchemeExpressive(DynamicScheme):
    """
    A Dynamic Color theme that is intentionally detached from the source color.
    """

    DEFAULT_SPEC_VERSION = '2021'
    DEFAULT_PLATFORM = 'phone'

    def __init__(
            self, 
            source_color_hct: Hct, 
            is_dark: bool, 
            contrast_level: float,
            spec_version: Optional[SpecVersion] = None,
            platform: Optional[Platform] = None
    ):
        """
        Construct a new SchemeExpressive instance.

        Args:
            source_color_hct: The source color of the theme as an HCT color.
            is_dark: Whether the scheme is in dark mode or light mode.
            contrast_level: The contrast level of the theme.
            spec_version: The specification version to use. Defaults to '2021'.
            platform: The platform to use. Defaults to 'phone'.
        """
        super().__init__({
            'source_color_hct': source_color_hct,
            'variant': Variant.EXPRESSIVE,
            'contrast_level': contrast_level,
            'is_dark': is_dark,
            'platform': platform or SchemeExpressive.DEFAULT_PLATFORM,
            'spec_version': spec_version or SchemeExpressive.DEFAULT_SPEC_VERSION,
        })


# hct/hct.py

"""
A color system built using CAM16 hue and chroma, and L* from
L*a*b*.

Using L* creates a link between the color system, contrast, and thus
accessibility. Contrast ratio depends on relative luminance, or Y in the XYZ
color space. L*, or perceptual luminance can be calculated from Y.

Unlike Y, L* is linear to human perception, allowing trivial creation of
accurate color tones.

Unlike contrast ratio, measuring contrast in L* is linear, and simple to
calculate. A difference of 40 in HCT tone guarantees a contrast ratio >= 3.0,
and a difference of 50 guarantees a contrast ratio >= 4.5.
"""

from PyMCUlib.utils import color_utils
from PyMCUlib.hct.cam16 import Cam16
from PyMCUlib.hct.hct_solver import HctSolver
from PyMCUlib.hct.viewing_conditions import ViewingConditions


class Hct:
    """
    HCT, hue, chroma, and tone. A color system that provides a perceptually
    accurate color measurement system that can also accurately render what colors
    will appear as in different lighting environments.
    """

    @classmethod
    def from_hct(cls, hue: float, chroma: float, tone: float) -> "Hct":
        """
        Create an HCT color from hue, chroma, and tone.
        
        Args:
            hue: 0 <= hue < 360; invalid values are corrected.
            chroma: 0 <= chroma < ?; Informally, colorfulness. The color
                returned may be lower than the requested chroma. Chroma has a different
                maximum for any given hue and tone.
            tone: 0 <= tone <= 100; invalid values are corrected.
        
        Returns:
            HCT representation of a color in default viewing conditions.
        """
        return cls(HctSolver.solve_to_int(hue, chroma, tone))

    @classmethod
    def from_int(cls, argb: int) -> "Hct":
        """
        Create an HCT color from an ARGB integer.
        
        Args:
            argb: ARGB representation of a color.
        
        Returns:
            HCT representation of a color in default viewing conditions
        """
        return cls(argb)

    def to_int(self) -> int:
        """
        Convert to ARGB integer representation.
        
        Returns:
            ARGB representation of the color
        """
        return self.argb

    @property
    def hue(self) -> float:
        """
        A number, in degrees, representing ex. red, orange, yellow, etc.
        Ranges from 0 <= hue < 360.
        """
        return self.internal_hue

    @hue.setter
    def hue(self, new_hue: float) -> None:
        """
        Set the hue of the color.
        
        Args:
            new_hue: 0 <= new_hue < 360; invalid values are corrected.
                Chroma may decrease because chroma has a different maximum for any given
                hue and tone.
        """
        self._set_internal_state(
            HctSolver.solve_to_int(
                new_hue,
                self.internal_chroma,
                self.internal_tone,
            ),
        )

    @property
    def chroma(self) -> float:
        """
        Get the chroma of the color.
        """
        return self.internal_chroma

    @chroma.setter
    def chroma(self, new_chroma: float) -> None:
        """
        Set the chroma of the color.
        
        Args:
            new_chroma: 0 <= new_chroma < ?
                Chroma may decrease because chroma has a different maximum for any given
                hue and tone.
        """
        self._set_internal_state(
            HctSolver.solve_to_int(
                self.internal_hue,
                new_chroma,
                self.internal_tone,
            ),
        )

    @property
    def tone(self) -> float:
        """
        Lightness. Ranges from 0 to 100.
        """
        return self.internal_tone

    @tone.setter
    def tone(self, new_tone: float) -> None:
        """
        Set the tone of the color.
        
        Args:
            new_tone: 0 <= new_tone <= 100; invalid values are corrected.
                Chroma may decrease because chroma has a different maximum for any given
                hue and tone.
        """
        self._set_internal_state(
            HctSolver.solve_to_int(
                self.internal_hue,
                self.internal_chroma,
                new_tone,
            ),
        )

    def set_value(self, property_name: str, value: float) -> None:
        """
        Sets a property of the Hct object.
        
        Args:
            property_name: String, property to set
            value: Value to set for the property
        """
        setattr(self, property_name, value)

    def __str__(self) -> str:
        """
        Returns a string representation of the color.
        
        Returns:
            String representation of the color
        """
        return f"HCT({int(self.hue)}, {int(self.chroma)}, {int(self.tone)})"

    @staticmethod
    def is_blue(hue: float) -> bool:
        """
        Check if the hue represents a blue color.
        
        Args:
            hue: Hue in degrees
            
        Returns:
            True if the hue is in the blue range
        """
        return hue >= 250 and hue < 270

    @staticmethod
    def is_yellow(hue: float) -> bool:
        """
        Check if the hue represents a yellow color.
        
        Args:
            hue: Hue in degrees
            
        Returns:
            True if the hue is in the yellow range
        """
        return hue >= 105 and hue < 125

    @staticmethod
    def is_cyan(hue: float) -> bool:
        """
        Check if the hue represents a cyan color.
        
        Args:
            hue: Hue in degrees
            
        Returns:
            True if the hue is in the cyan range
        """
        return hue >= 170 and hue < 207

    def __init__(self, argb: int) -> None:
        """
        Create an HCT color from an ARGB integer.
        
        Args:
            argb: ARGB representation of a color.
        """
        cam = Cam16.from_int(argb)
        self.internal_hue = cam.hue
        self.internal_chroma = cam.chroma
        self.internal_tone = color_utils.lstar_from_argb(argb)
        self.argb = argb

    def _set_internal_state(self, argb: int) -> None:
        """
        Update the internal state of the HCT object.
        
        Args:
            argb: ARGB representation of a color.
        """
        cam = Cam16.from_int(argb)
        self.internal_hue = cam.hue
        self.internal_chroma = cam.chroma
        self.internal_tone = color_utils.lstar_from_argb(argb)
        self.argb = argb

    def in_viewing_conditions(self, vc: ViewingConditions) -> "Hct":
        """
        Translates a color into different ViewingConditions.

        Colors change appearance. They look different with lights on versus off,
        the same color, as in hex code, on white looks different when on black.
        This is called color relativity, most famously explicated by Josef Albers
        in Interaction of Color.

        In color science, color appearance models can account for this and
        calculate the appearance of a color in different settings. HCT is based on
        CAM16, a color appearance model, and uses it to make these calculations.
        
        Args:
            vc: ViewingConditions object representing the viewing environment
            
        Returns:
            Hct object representing the color in the new viewing conditions
        
        See:
            ViewingConditions.make for parameters affecting color appearance.
        """
        # 1. Use CAM16 to find XYZ coordinates of color in specified VC.
        cam = Cam16.from_int(self.to_int())
        viewed_in_vc = cam.xyz_in_viewing_conditions(vc)

        # 2. Create CAM16 of those XYZ coordinates in default VC.
        recast_in_vc = Cam16.from_xyz_in_viewing_conditions(
            viewed_in_vc[0],
            viewed_in_vc[1],
            viewed_in_vc[2],
            ViewingConditions.make(),
        )

        # 3. Create HCT from:
        # - CAM16 using default VC with XYZ coordinates in specified VC.
        # - L* converted from Y in XYZ coordinates in specified VC.
        recast_hct = Hct.from_hct(
            recast_in_vc.hue,
            recast_in_vc.chroma,
            color_utils.lstar_from_y(viewed_in_vc[1]),
        )
        return recast_hct


# dislike/dislike_analyzer.py

"""
Check and/or fix universally disliked colors.
Color science studies of color preference indicate universal distaste for
dark yellow-greens, and also show this is correlated to distate for
biological waste and rotting food.

See Palmer and Schloss, 2010 or Schloss and Palmer's Chapter 21 in Handbook
of Color Psychology (2015).
"""

from PyMCUlib.hct.hct import Hct


class DislikeAnalyzer:
    """
    Check and/or fix universally disliked colors.
    Color science studies of color preference indicate universal distaste for
    dark yellow-greens, and also show this is correlated to distate for
    biological waste and rotting food.

    See Palmer and Schloss, 2010 or Schloss and Palmer's Chapter 21 in Handbook
    of Color Psychology (2015).
    """

    @staticmethod
    def is_disliked(hct: Hct) -> bool:
        """
        Returns true if a color is disliked.

        Args:
            hct: A color to be judged.

        Returns:
            Whether the color is disliked.

        Disliked is defined as a dark yellow-green that is not neutral.
        """
        hue_passes = round(hct.hue) >= 90.0 and round(hct.hue) <= 111.0
        chroma_passes = round(hct.chroma) > 16.0
        tone_passes = round(hct.tone) < 65.0

        return hue_passes and chroma_passes and tone_passes

    @staticmethod
    def fix_if_disliked(hct: Hct) -> Hct:
        """
        If a color is disliked, lighten it to make it likable.

        Args:
            hct: A color to be judged.

        Returns:
            A new color if the original color is disliked, or the original
            color if it is acceptable.
        """
        if DislikeAnalyzer.is_disliked(hct):
            return Hct.from_hct(
                hct.hue,
                hct.chroma,
                70.0,
            )

        return hct


# hct/viewing_conditions.py

import math
from typing import List, Optional
from PyMCUlib.utils import color_utils
from PyMCUlib.utils import math_utils

class ViewingConditions:
    """
    In traditional color spaces, a color can be identified solely by the
    observer's measurement of the color. Color appearance models such as CAM16
    also use information about the environment where the color was
    observed, known as the viewing conditions.

    For example, white under the traditional assumption of a midday sun white
    point is accurately measured as a slightly chromatic blue by CAM16. (roughly,
    hue 203, chroma 3, lightness 100)

    This class caches intermediate values of the CAM16 conversion process that
    depend only on viewing conditions, enabling speed ups.
    """

    @classmethod
    def make(
        cls,
        white_point: Optional[List[float]] = None,
        adapting_luminance: float = (200.0 / math.pi) * color_utils.y_from_lstar(50.0) / 100.0,
        background_lstar: float = 50.0,
        surround: float = 2.0,
        discounting_illuminant: bool = False,
    ) -> "ViewingConditions":
        """
        Create ViewingConditions from a simple, physically relevant, set of
        parameters.

        Args:
            white_point: White point, measured in the XYZ color space.
                default = D65, or sunny day afternoon
            adapting_luminance: The luminance of the adapting field. Informally,
                how bright it is in the room where the color is viewed. Can be
                calculated from lux by multiplying lux by 0.0586. default = 11.72,
                or 200 lux.
            background_lstar: The lightness of the area surrounding the color.
                measured by L* in L*a*b*. default = 50.0
            surround: A general description of the lighting surrounding the
                color. 0 is pitch dark, like watching a movie in a theater. 1.0 is a
                dimly light room, like watching TV at home at night. 2.0 means there
                is no difference between the lighting on the color and around it.
                default = 2.0
            discounting_illuminant: Whether the eye accounts for the tint of the
                ambient lighting, such as knowing an apple is still red in green light.
                default = false, the eye does not perform this process on
                self-luminous objects like displays.

        Returns:
            ViewingConditions instance
        """
        if white_point is None:
            white_point = color_utils.white_point_d65()

        xyz = white_point
        r_w = xyz[0] * 0.401288 + xyz[1] * 0.650173 + xyz[2] * -0.051461
        g_w = xyz[0] * -0.250268 + xyz[1] * 1.204414 + xyz[2] * 0.045854
        b_w = xyz[0] * -0.002079 + xyz[1] * 0.048952 + xyz[2] * 0.953127

        f = 0.8 + surround / 10.0
        c = (math_utils.lerp(0.59, 0.69, (f - 0.9) * 10.0) if f >= 0.9 
             else math_utils.lerp(0.525, 0.59, (f - 0.8) * 10.0))

        d = (1.0 if discounting_illuminant 
             else f * (1.0 - (1.0 / 3.6) * math.exp((-adapting_luminance - 42.0) / 92.0)))
        d = 1.0 if d > 1.0 else (0.0 if d < 0.0 else d)

        nc = f
        rgb_d = [
            d * (100.0 / r_w) + 1.0 - d,
            d * (100.0 / g_w) + 1.0 - d,
            d * (100.0 / b_w) + 1.0 - d,
        ]

        k = 1.0 / (5.0 * adapting_luminance + 1.0)
        k4 = k * k * k * k
        k4F = 1.0 - k4
        fl = (k4 * adapting_luminance +
              0.1 * k4F * k4F * math.pow(5.0 * adapting_luminance, 1/3))

        n = color_utils.y_from_lstar(background_lstar) / white_point[1]
        z = 1.48 + math.sqrt(n)
        nbb = 0.725 / math.pow(n, 0.2)
        ncb = nbb

        rgb_a_factors = [
            math.pow((fl * rgb_d[0] * r_w) / 100.0, 0.42),
            math.pow((fl * rgb_d[1] * g_w) / 100.0, 0.42),
            math.pow((fl * rgb_d[2] * b_w) / 100.0, 0.42),
        ]

        rgb_a = [
            (400.0 * rgb_a_factors[0]) / (rgb_a_factors[0] + 27.13),
            (400.0 * rgb_a_factors[1]) / (rgb_a_factors[1] + 27.13),
            (400.0 * rgb_a_factors[2]) / (rgb_a_factors[2] + 27.13),
        ]

        aw = (2.0 * rgb_a[0] + rgb_a[1] + 0.05 * rgb_a[2]) * nbb

        return cls(n, aw, nbb, ncb, c, nc, rgb_d, fl, math.pow(fl, 0.25), z)

    def __init__(self, n: float, aw: float, nbb: float, ncb: float, c: float, nc: float, rgb_d: List[float], fl: float, fl_root: float, z: float) -> None:
        """
        Parameters are intermediate values of the CAM16 conversion process. Their
        names are shorthand for technical color science terminology, this class
        would not benefit from documenting them individually. A brief overview
        is available in the CAM16 specification, and a complete overview requires
        a color science textbook, such as Fairchild's Color Appearance Models.
        """
        self.n = n
        self.aw = aw
        self.nbb = nbb
        self.ncb = ncb
        self.c = c
        self.nc = nc
        self.rgb_d = rgb_d
        self.fl = fl
        self.fl_root = fl_root
        self.z = z


# Set the default viewing conditions
ViewingConditions.DEFAULT = ViewingConditions.make()


# palettes/__init__.py

# --- tonal_palette ---
from PyMCUlib.palettes.tonal_palette import TonalPalette, KeyColor

# --- core_palette ---
from PyMCUlib.palettes.core_palette import CorePalette

# --- core_palettes ---
from PyMCUlib.palettes.core_palettes import CorePalettes

__all__ = [
    "TonalPalette",
    "KeyColor",
    "CorePalette",
    "CorePalettes",
]


# dynamiccolor/material_dynamic_colors.py

from typing import List, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from PyMCUlib.dynamiccolor.dynamic_scheme import DynamicScheme

from PyMCUlib.dynamiccolor.color_spec_2025 import ColorSpecDelegateImpl2025
from PyMCUlib.dynamiccolor.dynamic_color import DynamicColor


class MaterialDynamicColors:
    """
    DynamicColors for the colors in the Material Design system.
    """
    content_accent_tone_delta = 15.0
    
    _color_spec = ColorSpecDelegateImpl2025()
    
    def highest_surface(self, s: 'DynamicScheme') -> DynamicColor:
        """
        Returns the highest surface color for the given scheme.
        
        Args:
            s: The dynamic scheme to get the highest surface for
            
        Returns:
            The highest surface dynamic color
        """
        return MaterialDynamicColors._color_spec.highest_surface(s)
    
    ################################################################
    # Main Palettes                                              #
    ################################################################
    
    def primary_palette_key_color(self) -> DynamicColor:
        """
        Returns the primary palette key color.
        
        Returns:
            The primary palette key dynamic color
        """
        return MaterialDynamicColors._color_spec.primary_palette_key_color()
    
    def secondary_palette_key_color(self) -> DynamicColor:
        """
        Returns the secondary palette key color.
        
        Returns:
            The secondary palette key dynamic color
        """
        return MaterialDynamicColors._color_spec.secondary_palette_key_color()
    
    def tertiary_palette_key_color(self) -> DynamicColor:
        """
        Returns the tertiary palette key color.
        
        Returns:
            The tertiary palette key dynamic color
        """
        return MaterialDynamicColors._color_spec.tertiary_palette_key_color()
    
    def neutral_palette_key_color(self) -> DynamicColor:
        """
        Returns the neutral palette key color.
        
        Returns:
            The neutral palette key dynamic color
        """
        return MaterialDynamicColors._color_spec.neutral_palette_key_color()
    
    def neutral_variant_palette_key_color(self) -> DynamicColor:
        """
        Returns the neutral variant palette key color.
        
        Returns:
            The neutral variant palette key dynamic color
        """
        return MaterialDynamicColors._color_spec.neutral_variant_palette_key_color()
    
    def error_palette_key_color(self) -> DynamicColor:
        """
        Returns the error palette key color.
        
        Returns:
            The error palette key dynamic color
        """
        return MaterialDynamicColors._color_spec.error_palette_key_color()
    
    ################################################################
    # Surfaces [S]                                               #
    ################################################################
    
    def background(self) -> DynamicColor:
        """
        Returns the background color.
        
        Returns:
            The background dynamic color
        """
        return MaterialDynamicColors._color_spec.background()
    
    def on_background(self) -> DynamicColor:
        """
        Returns the on background color.
        
        Returns:
            The on background dynamic color
        """
        return MaterialDynamicColors._color_spec.on_background()
    
    def surface(self) -> DynamicColor:
        """
        Returns the surface color.
        
        Returns:
            The surface dynamic color
        """
        return MaterialDynamicColors._color_spec.surface()
    
    def surface_dim(self) -> DynamicColor:
        """
        Returns the surface dim color.
        
        Returns:
            The surface dim dynamic color
        """
        return MaterialDynamicColors._color_spec.surface_dim()
    
    def surface_bright(self) -> DynamicColor:
        """
        Returns the surface bright color.
        
        Returns:
            The surface bright dynamic color
        """
        return MaterialDynamicColors._color_spec.surface_bright()
    
    def surface_container_lowest(self) -> DynamicColor:
        """
        Returns the surface container lowest color.
        
        Returns:
            The surface container lowest dynamic color
        """
        return MaterialDynamicColors._color_spec.surface_container_lowest()
    
    def surface_container_low(self) -> DynamicColor:
        """
        Returns the surface container low color.
        
        Returns:
            The surface container low dynamic color
        """
        return MaterialDynamicColors._color_spec.surface_container_low()
    
    def surface_container(self) -> DynamicColor:
        """
        Returns the surface container color.
        
        Returns:
            The surface container dynamic color
        """
        return MaterialDynamicColors._color_spec.surface_container()
    
    def surface_container_high(self) -> DynamicColor:
        """
        Returns the surface container high color.
        
        Returns:
            The surface container high dynamic color
        """
        return MaterialDynamicColors._color_spec.surface_container_high()
    
    def surface_container_highest(self) -> DynamicColor:
        """
        Returns the surface container highest color.
        
        Returns:
            The surface container highest dynamic color
        """
        return MaterialDynamicColors._color_spec.surface_container_highest()
    
    def on_surface(self) -> DynamicColor:
        """
        Returns the on surface color.
        
        Returns:
            The on surface dynamic color
        """
        return MaterialDynamicColors._color_spec.on_surface()
    
    def surface_variant(self) -> DynamicColor:
        """
        Returns the surface variant color.
        
        Returns:
            The surface variant dynamic color
        """
        return MaterialDynamicColors._color_spec.surface_variant()
    
    def on_surface_variant(self) -> DynamicColor:
        """
        Returns the on surface variant color.
        
        Returns:
            The on surface variant dynamic color
        """
        return MaterialDynamicColors._color_spec.on_surface_variant()
    
    def outline(self) -> DynamicColor:
        """
        Returns the outline color.
        
        Returns:
            The outline dynamic color
        """
        return MaterialDynamicColors._color_spec.outline()
    
    def outline_variant(self) -> DynamicColor:
        """
        Returns the outline variant color.
        
        Returns:
            The outline variant dynamic color
        """
        return MaterialDynamicColors._color_spec.outline_variant()
    
    def inverse_surface(self) -> DynamicColor:
        """
        Returns the inverse surface color.
        
        Returns:
            The inverse surface dynamic color
        """
        return MaterialDynamicColors._color_spec.inverse_surface()
    
    def inverse_on_surface(self) -> DynamicColor:
        """
        Returns the inverse on surface color.
        
        Returns:
            The inverse on surface dynamic color
        """
        return MaterialDynamicColors._color_spec.inverse_on_surface()
    
    def shadow(self) -> DynamicColor:
        """
        Returns the shadow color.
        
        Returns:
            The shadow dynamic color
        """
        return MaterialDynamicColors._color_spec.shadow()
    
    def scrim(self) -> DynamicColor:
        """
        Returns the scrim color.
        
        Returns:
            The scrim dynamic color
        """
        return MaterialDynamicColors._color_spec.scrim()
    
    def surface_tint(self) -> DynamicColor:
        """
        Returns the surface tint color.
        
        Returns:
            The surface tint dynamic color
        """
        return MaterialDynamicColors._color_spec.surface_tint()
    
    ################################################################
    # Primaries [P]                                              #
    ################################################################
    
    def primary(self) -> DynamicColor:
        """
        Returns the primary color.
        
        Returns:
            The primary dynamic color
        """
        return MaterialDynamicColors._color_spec.primary()
    
    def primary_dim(self) -> Optional[DynamicColor]:
        """
        Returns the primary dim color.
        
        Returns:
            The primary dim dynamic color or None if not available
        """
        return MaterialDynamicColors._color_spec.primary_dim()
    
    def on_primary(self) -> DynamicColor:
        """
        Returns the on primary color.
        
        Returns:
            The on primary dynamic color
        """
        return MaterialDynamicColors._color_spec.on_primary()
    
    def primary_container(self) -> DynamicColor:
        """
        Returns the primary container color.
        
        Returns:
            The primary container dynamic color
        """
        return MaterialDynamicColors._color_spec.primary_container()
    
    def on_primary_container(self) -> DynamicColor:
        """
        Returns the on primary container color.
        
        Returns:
            The on primary container dynamic color
        """
        return MaterialDynamicColors._color_spec.on_primary_container()
    
    def inverse_primary(self) -> DynamicColor:
        """
        Returns the inverse primary color.
        
        Returns:
            The inverse primary dynamic color
        """
        return MaterialDynamicColors._color_spec.inverse_primary()
    
    #################################################################
    # Primary Fixed [PF]                                          #
    #################################################################
    
    def primary_fixed(self) -> DynamicColor:
        """
        Returns the primary fixed color.
        
        Returns:
            The primary fixed dynamic color
        """
        return MaterialDynamicColors._color_spec.primary_fixed()
    
    def primary_fixed_dim(self) -> DynamicColor:
        """
        Returns the primary fixed dim color.
        
        Returns:
            The primary fixed dim dynamic color
        """
        return MaterialDynamicColors._color_spec.primary_fixed_dim()
    
    def on_primary_fixed(self) -> DynamicColor:
        """
        Returns the on primary fixed color.
        
        Returns:
            The on primary fixed dynamic color
        """
        return MaterialDynamicColors._color_spec.on_primary_fixed()
    
    def on_primary_fixed_variant(self) -> DynamicColor:
        """
        Returns the on primary fixed variant color.
        
        Returns:
            The on primary fixed variant dynamic color
        """
        return MaterialDynamicColors._color_spec.on_primary_fixed_variant()
    
    ################################################################
    # Secondaries [Q]                                            #
    ################################################################
    
    def secondary(self) -> DynamicColor:
        """
        Returns the secondary color.
        
        Returns:
            The secondary dynamic color
        """
        return MaterialDynamicColors._color_spec.secondary()
    
    def secondary_dim(self) -> Optional[DynamicColor]:
        """
        Returns the secondary dim color.
        
        Returns:
            The secondary dim dynamic color or None if not available
        """
        return MaterialDynamicColors._color_spec.secondary_dim()
    
    def on_secondary(self) -> DynamicColor:
        """
        Returns the on secondary color.
        
        Returns:
            The on secondary dynamic color
        """
        return MaterialDynamicColors._color_spec.on_secondary()
    
    def secondary_container(self) -> DynamicColor:
        """
        Returns the secondary container color.
        
        Returns:
            The secondary container dynamic color
        """
        return MaterialDynamicColors._color_spec.secondary_container()
    
    def on_secondary_container(self) -> DynamicColor:
        """
        Returns the on secondary container color.
        
        Returns:
            The on secondary container dynamic color
        """
        return MaterialDynamicColors._color_spec.on_secondary_container()
    
    #################################################################
    # Secondary Fixed [QF]                                        #
    #################################################################
    
    def secondary_fixed(self) -> DynamicColor:
        """
        Returns the secondary fixed color.
        
        Returns:
            The secondary fixed dynamic color
        """
        return MaterialDynamicColors._color_spec.secondary_fixed()
    
    def secondary_fixed_dim(self) -> DynamicColor:
        """
        Returns the secondary fixed dim color.
        
        Returns:
            The secondary fixed dim dynamic color
        """
        return MaterialDynamicColors._color_spec.secondary_fixed_dim()
    
    def on_secondary_fixed(self) -> DynamicColor:
        """
        Returns the on secondary fixed color.
        
        Returns:
            The on secondary fixed dynamic color
        """
        return MaterialDynamicColors._color_spec.on_secondary_fixed()
    
    def on_secondary_fixed_variant(self) -> DynamicColor:
        """
        Returns the on secondary fixed variant color.
        
        Returns:
            The on secondary fixed variant dynamic color
        """
        return MaterialDynamicColors._color_spec.on_secondary_fixed_variant()
    
    ################################################################
    # Tertiaries [T]                                             #
    ################################################################
    
    def tertiary(self) -> DynamicColor:
        """
        Returns the tertiary color.
        
        Returns:
            The tertiary dynamic color
        """
        return MaterialDynamicColors._color_spec.tertiary()
    
    def tertiary_dim(self) -> Optional[DynamicColor]:
        """
        Returns the tertiary dim color.
        
        Returns:
            The tertiary dim dynamic color or None if not available
        """
        return MaterialDynamicColors._color_spec.tertiary_dim()
    
    def on_tertiary(self) -> DynamicColor:
        """
        Returns the on tertiary color.
        
        Returns:
            The on tertiary dynamic color
        """
        return MaterialDynamicColors._color_spec.on_tertiary()
    
    def tertiary_container(self) -> DynamicColor:
        """
        Returns the tertiary container color.
        
        Returns:
            The tertiary container dynamic color
        """
        return MaterialDynamicColors._color_spec.tertiary_container()
    
    def on_tertiary_container(self) -> DynamicColor:
        """
        Returns the on tertiary container color.
        
        Returns:
            The on tertiary container dynamic color
        """
        return MaterialDynamicColors._color_spec.on_tertiary_container()
    
    #################################################################
    # Tertiary Fixed [TF]                                         #
    #################################################################
    
    def tertiary_fixed(self) -> DynamicColor:
        """
        Returns the tertiary fixed color.
        
        Returns:
            The tertiary fixed dynamic color
        """
        return MaterialDynamicColors._color_spec.tertiary_fixed()
    
    def tertiary_fixed_dim(self) -> DynamicColor:
        """
        Returns the tertiary fixed dim color.
        
        Returns:
            The tertiary fixed dim dynamic color
        """
        return MaterialDynamicColors._color_spec.tertiary_fixed_dim()
    
    def on_tertiary_fixed(self) -> DynamicColor:
        """
        Returns the on tertiary fixed color.
        
        Returns:
            The on tertiary fixed dynamic color
        """
        return MaterialDynamicColors._color_spec.on_tertiary_fixed()
    
    def on_tertiary_fixed_variant(self) -> DynamicColor:
        """
        Returns the on tertiary fixed variant color.
        
        Returns:
            The on tertiary fixed variant dynamic color
        """
        return MaterialDynamicColors._color_spec.on_tertiary_fixed_variant()
    
    ################################################################
    # Errors [E]                                                 #
    ################################################################
    
    def error(self) -> DynamicColor:
        """
        Returns the error color.
        
        Returns:
            The error dynamic color
        """
        return MaterialDynamicColors._color_spec.error()
    
    def error_dim(self) -> Optional[DynamicColor]:
        """
        Returns the error dim color.
        
        Returns:
            The error dim dynamic color or None if not available
        """
        return MaterialDynamicColors._color_spec.error_dim()
    
    def on_error(self) -> DynamicColor:
        """
        Returns the on error color.
        
        Returns:
            The on error dynamic color
        """
        return MaterialDynamicColors._color_spec.on_error()
    
    def error_container(self) -> DynamicColor:
        """
        Returns the error container color.
        
        Returns:
            The error container dynamic color
        """
        return MaterialDynamicColors._color_spec.error_container()
    
    def on_error_container(self) -> DynamicColor:
        """
        Returns the on error container color.
        
        Returns:
            The on error container dynamic color
        """
        return MaterialDynamicColors._color_spec.on_error_container()
    
    ################################################################
    # All Colors                                                 #
    ################################################################
    
    @property
    def all_colors(self) -> List[DynamicColor]:
        """
        List of all available dynamic colors.
        
        Returns:
            List of all non-undefined dynamic colors
        """
        colors = [
            self.background(),
            self.on_background(),
            self.surface(),
            self.surface_dim(),
            self.surface_bright(),
            self.surface_container_lowest(),
            self.surface_container_low(),
            self.surface_container(),
            self.surface_container_high(),
            self.surface_container_highest(),
            self.on_surface(),
            self.on_surface_variant(),
            self.outline(),
            self.outline_variant(),
            self.inverse_surface(),
            self.inverse_on_surface(),
            self.primary(),
            self.primary_dim(),
            self.on_primary(),
            self.primary_container(),
            self.on_primary_container(),
            self.primary_fixed(),
            self.primary_fixed_dim(),
            self.on_primary_fixed(),
            self.on_primary_fixed_variant(),
            self.inverse_primary(),
            self.secondary(),
            self.secondary_dim(),
            self.on_secondary(),
            self.secondary_container(),
            self.on_secondary_container(),
            self.secondary_fixed(),
            self.secondary_fixed_dim(),
            self.on_secondary_fixed(),
            self.on_secondary_fixed_variant(),
            self.tertiary(),
            self.tertiary_dim(),
            self.on_tertiary(),
            self.tertiary_container(),
            self.on_tertiary_container(),
            self.tertiary_fixed(),
            self.tertiary_fixed_dim(),
            self.on_tertiary_fixed(),
            self.on_tertiary_fixed_variant(),
            self.error(),
            self.error_dim(),
            self.on_error(),
            self.error_container(),
            self.on_error_container(),
        ]
        return [c for c in colors if c is not None]


# Static properties are deprecated, but provided for backwards compatibility
# These are implemented as class attributes in Python
# @deprecated Use highest_surface() instead.
MaterialDynamicColors.highest_surface = MaterialDynamicColors._color_spec.highest_surface

# @deprecated Use primaryPaletteKeyColor() instead.
MaterialDynamicColors.primary_palette_key_color = MaterialDynamicColors._color_spec.primary_palette_key_color()

# @deprecated Use secondaryPaletteKeyColor() instead.
MaterialDynamicColors.secondary_palette_key_color = MaterialDynamicColors._color_spec.secondary_palette_key_color()

# @deprecated Use tertiaryPaletteKeyColor() instead.
MaterialDynamicColors.tertiary_palette_key_color = MaterialDynamicColors._color_spec.tertiary_palette_key_color()

# @deprecated Use neutralPaletteKeyColor() instead.
MaterialDynamicColors.neutral_palette_key_color = MaterialDynamicColors._color_spec.neutral_palette_key_color()

# @deprecated Use neutralVariantPaletteKeyColor() instead.
MaterialDynamicColors.neutral_variant_palette_key_color = MaterialDynamicColors._color_spec.neutral_variant_palette_key_color()

# @deprecated Use background() instead.
MaterialDynamicColors.background = MaterialDynamicColors._color_spec.background()

# @deprecated Use onBackground() instead.
MaterialDynamicColors.on_background = MaterialDynamicColors._color_spec.on_background()

# @deprecated Use surface() instead.
MaterialDynamicColors.surface = MaterialDynamicColors._color_spec.surface()

# @deprecated Use surfaceDim() instead.
MaterialDynamicColors.surface_dim = MaterialDynamicColors._color_spec.surface_dim()

# @deprecated Use surfaceBright() instead.
MaterialDynamicColors.surface_bright = MaterialDynamicColors._color_spec.surface_bright()

# @deprecated Use surfaceContainerLowest() instead.
MaterialDynamicColors.surface_container_lowest = MaterialDynamicColors._color_spec.surface_container_lowest()

# @deprecated Use surfaceContainerLow() instead.
MaterialDynamicColors.surface_container_low = MaterialDynamicColors._color_spec.surface_container_low()

# @deprecated Use surfaceContainer() instead.
MaterialDynamicColors.surface_container = MaterialDynamicColors._color_spec.surface_container()

# @deprecated Use surfaceContainerHigh() instead.
MaterialDynamicColors.surface_container_high = MaterialDynamicColors._color_spec.surface_container_high()

# @deprecated Use surfaceContainerHighest() instead.
MaterialDynamicColors.surface_container_highest = MaterialDynamicColors._color_spec.surface_container_highest()

# @deprecated Use onSurface() instead.
MaterialDynamicColors.on_surface = MaterialDynamicColors._color_spec.on_surface()

# @deprecated Use surfaceVariant() instead.
MaterialDynamicColors.surface_variant = MaterialDynamicColors._color_spec.surface_variant()

# @deprecated Use onSurfaceVariant() instead.
MaterialDynamicColors.on_surface_variant = MaterialDynamicColors._color_spec.on_surface_variant()

# @deprecated Use inverseSurface() instead.
MaterialDynamicColors.inverse_surface = MaterialDynamicColors._color_spec.inverse_surface()

# @deprecated Use inverseOnSurface() instead.
MaterialDynamicColors.inverse_on_surface = MaterialDynamicColors._color_spec.inverse_on_surface()

# @deprecated Use outline() instead.
MaterialDynamicColors.outline = MaterialDynamicColors._color_spec.outline()

# @deprecated Use outlineVariant() instead.
MaterialDynamicColors.outline_variant = MaterialDynamicColors._color_spec.outline_variant()

# @deprecated Use shadow() instead.
MaterialDynamicColors.shadow = MaterialDynamicColors._color_spec.shadow()

# @deprecated Use scrim() instead.
MaterialDynamicColors.scrim = MaterialDynamicColors._color_spec.scrim()

# @deprecated Use surfaceTint() instead.
MaterialDynamicColors.surface_tint = MaterialDynamicColors._color_spec.surface_tint()

# @deprecated Use primary() instead.
MaterialDynamicColors.primary = MaterialDynamicColors._color_spec.primary()

# @deprecated Use onPrimary() instead.
MaterialDynamicColors.on_primary = MaterialDynamicColors._color_spec.on_primary()

# @deprecated Use primaryContainer() instead.
MaterialDynamicColors.primary_container = MaterialDynamicColors._color_spec.primary_container()

# @deprecated Use onPrimaryContainer() instead.
MaterialDynamicColors.on_primary_container = MaterialDynamicColors._color_spec.on_primary_container()

# @deprecated Use inversePrimary() instead.
MaterialDynamicColors.inverse_primary = MaterialDynamicColors._color_spec.inverse_primary()

# @deprecated Use secondary() instead.
MaterialDynamicColors.secondary = MaterialDynamicColors._color_spec.secondary()

# @deprecated Use onSecondary() instead.
MaterialDynamicColors.on_secondary = MaterialDynamicColors._color_spec.on_secondary()

# @deprecated Use secondaryContainer() instead.
MaterialDynamicColors.secondary_container = MaterialDynamicColors._color_spec.secondary_container()

# @deprecated Use onSecondaryContainer() instead.
MaterialDynamicColors.on_secondary_container = MaterialDynamicColors._color_spec.on_secondary_container()

# @deprecated Use tertiary() instead.
MaterialDynamicColors.tertiary = MaterialDynamicColors._color_spec.tertiary()

# @deprecated Use onTertiary() instead.
MaterialDynamicColors.on_tertiary = MaterialDynamicColors._color_spec.on_tertiary()

# @deprecated Use tertiaryContainer() instead.
MaterialDynamicColors.tertiary_container = MaterialDynamicColors._color_spec.tertiary_container()

# @deprecated Use onTertiaryContainer() instead.
MaterialDynamicColors.on_tertiary_container = MaterialDynamicColors._color_spec.on_tertiary_container()

# @deprecated Use error() instead.
MaterialDynamicColors.error = MaterialDynamicColors._color_spec.error()

# @deprecated Use onError() instead.
MaterialDynamicColors.on_error = MaterialDynamicColors._color_spec.on_error()

# @deprecated Use errorContainer() instead.
MaterialDynamicColors.error_container = MaterialDynamicColors._color_spec.error_container()

# @deprecated Use onErrorContainer() instead.
MaterialDynamicColors.on_error_container = MaterialDynamicColors._color_spec.on_error_container()

# @deprecated Use primaryFixed() instead.
MaterialDynamicColors.primary_fixed = MaterialDynamicColors._color_spec.primary_fixed()

# @deprecated Use primaryFixedDim() instead.
MaterialDynamicColors.primary_fixed_dim = MaterialDynamicColors._color_spec.primary_fixed_dim()

# @deprecated Use onPrimaryFixed() instead.
MaterialDynamicColors.on_primary_fixed = MaterialDynamicColors._color_spec.on_primary_fixed()

# @deprecated Use onPrimaryFixedVariant() instead.
MaterialDynamicColors.on_primary_fixed_variant = MaterialDynamicColors._color_spec.on_primary_fixed_variant()

# @deprecated Use secondaryFixed() instead.
MaterialDynamicColors.secondary_fixed = MaterialDynamicColors._color_spec.secondary_fixed()

# @deprecated Use secondaryFixedDim() instead.
MaterialDynamicColors.secondary_fixed_dim = MaterialDynamicColors._color_spec.secondary_fixed_dim()

# @deprecated Use onSecondaryFixed() instead.
MaterialDynamicColors.on_secondary_fixed = MaterialDynamicColors._color_spec.on_secondary_fixed()

# @deprecated Use onSecondaryFixedVariant() instead.
MaterialDynamicColors.on_secondary_fixed_variant = MaterialDynamicColors._color_spec.on_secondary_fixed_variant()

# @deprecated Use tertiaryFixed() instead.
MaterialDynamicColors.tertiary_fixed = MaterialDynamicColors._color_spec.tertiary_fixed()

# @deprecated Use tertiaryFixedDim() instead.
MaterialDynamicColors.tertiary_fixed_dim = MaterialDynamicColors._color_spec.tertiary_fixed_dim()

# @deprecated Use onTertiaryFixed() instead.
MaterialDynamicColors.on_tertiary_fixed = MaterialDynamicColors._color_spec.on_tertiary_fixed()

# @deprecated Use onTertiaryFixedVariant() instead.
MaterialDynamicColors.on_tertiary_fixed_variant = MaterialDynamicColors._color_spec.on_tertiary_fixed_variant()


# hct/cam16.py

import math
from PyMCUlib.utils import color_utils
from PyMCUlib.utils import math_utils
from PyMCUlib.hct.viewing_conditions import ViewingConditions

class Cam16:
    """
    CAM16, a color appearance model. Colors are not just defined by their hex
    code, but rather, a hex code and viewing conditions.

    CAM16 instances also have coordinates in the CAM16-UCS space, called J*, a*,
    b*, or jstar, astar, bstar in code. CAM16-UCS is included in the CAM16
    specification, and should be used when measuring distances between colors.

    In traditional color spaces, a color can be identified solely by the
    observer's measurement of the color. Color appearance models such as CAM16
    also use information about the environment where the color was
    observed, known as the viewing conditions.

    For example, white under the traditional assumption of a midday sun white
    point is accurately measured as a slightly chromatic blue by CAM16. (roughly,
    hue 203, chroma 3, lightness 100)
    """

    def __init__(
       self,
       hue: float, chroma: float, j: float, q: float,
       m: float, s: float, jstar: float, astar: float, bstar: float
    ) -> None:
        """
        All of the CAM16 dimensions can be calculated from 3 of the dimensions, in
        the following combinations:
            -  {j or q} and {c, m, or s} and hue
            - jstar, astar, bstar
        Prefer using a static method that constructs from 3 of those dimensions.
        This constructor is intended for those methods to use to return all
        possible dimensions.

        Args:
            hue: Hue in CAM16
            chroma: Chroma in CAM16, informally, colorfulness / color intensity. 
                    Like saturation in HSL, except perceptually accurate.
            j: Lightness in CAM16
            q: Brightness in CAM16; ratio of lightness to white point's lightness
            m: Colorfulness in CAM16
            s: Saturation in CAM16; ratio of chroma to white point's chroma
            jstar: CAM16-UCS J coordinate
            astar: CAM16-UCS a coordinate
            bstar: CAM16-UCS b coordinate
        """
        self.hue = hue
        self.chroma = chroma
        self.j = j
        self.q = q
        self.m = m
        self.s = s
        self.jstar = jstar
        self.astar = astar
        self.bstar = bstar
    
    def distance(self, other: 'Cam16') -> float:
        """
        CAM16 instances also have coordinates in the CAM16-UCS space, called J*,
        a*, b*, or jstar, astar, bstar in code. CAM16-UCS is included in the CAM16
        specification, and is used to measure distances between colors.

        Args:
            other: Another CAM16 instance

        Returns:
            The distance between the colors in CAM16-UCS space.
        """
        d_j = self.jstar - other.jstar
        d_a = self.astar - other.astar
        d_b = self.bstar - other.bstar
        d_e_prime = math.sqrt(d_j * d_j + d_a * d_a + d_b * d_b)
        d_e = 1.41 * math.pow(d_e_prime, 0.63)
        return d_e
    
    @classmethod
    def from_int(cls, argb: int) -> 'Cam16':
        """
        Creates a CAM16 color from an ARGB integer.

        Args:
            argb: ARGB representation of a color.

        Returns:
            CAM16 color, assuming the color was viewed in default viewing conditions.
        """
        return cls.from_int_in_viewing_conditions(argb, ViewingConditions.DEFAULT)
    
    @classmethod
    def from_int_in_viewing_conditions(
        cls,
        argb: int,
        viewing_conditions: ViewingConditions
    ) -> 'Cam16':
        """
        Creates a CAM16 color from an ARGB integer and viewing conditions.

        Args:
            argb: ARGB representation of a color.
            viewing_conditions: Information about the environment where the color
                was observed.

        Returns:
            CAM16 color
        """
        red = (argb & 0x00ff0000) >> 16
        green = (argb & 0x0000ff00) >> 8
        blue = (argb & 0x000000ff)
        red_l = color_utils.linearized(red)
        green_l = color_utils.linearized(green)
        blue_l = color_utils.linearized(blue)
        x = 0.41233895 * red_l + 0.35762064 * green_l + 0.18051042 * blue_l
        y = 0.2126 * red_l + 0.7152 * green_l + 0.0722 * blue_l
        z = 0.01932141 * red_l + 0.11916382 * green_l + 0.95034478 * blue_l
        
        r_c = 0.401288 * x + 0.650173 * y - 0.051461 * z
        g_c = -0.250268 * x + 1.204414 * y + 0.045854 * z
        b_c = -0.002079 * x + 0.048952 * y + 0.953127 * z
        
        r_d = viewing_conditions.rgb_d[0] * r_c
        g_d = viewing_conditions.rgb_d[1] * g_c
        b_d = viewing_conditions.rgb_d[2] * b_c
        
        r_af = math.pow((viewing_conditions.fl * abs(r_d)) / 100.0, 0.42)
        g_af = math.pow((viewing_conditions.fl * abs(g_d)) / 100.0, 0.42)
        b_af = math.pow((viewing_conditions.fl * abs(b_d)) / 100.0, 0.42)
        
        r_a = (math_utils.signum(r_d) * 400.0 * r_af) / (r_af + 27.13)
        g_a = (math_utils.signum(g_d) * 400.0 * g_af) / (g_af + 27.13)
        b_a = (math_utils.signum(b_d) * 400.0 * b_af) / (b_af + 27.13)
        
        a = (11.0 * r_a + -12.0 * g_a + b_a) / 11.0
        b = (r_a + g_a - 2.0 * b_a) / 9.0
        u = (20.0 * r_a + 20.0 * g_a + 21.0 * b_a) / 20.0
        p2 = (40.0 * r_a + 20.0 * g_a + b_a) / 20.0
        
        atan2 = math.atan2(b, a)
        atan_degrees = (atan2 * 180.0) / math.pi
        hue = atan_degrees if atan_degrees >= 0 else atan_degrees + 360.0
        hue = hue if hue < 360 else hue - 360.0
        hue_radians = (hue * math.pi) / 180.0
        
        ac = p2 * viewing_conditions.nbb
        
        j = 100.0 * math.pow(
            ac / viewing_conditions.aw,
            viewing_conditions.c * viewing_conditions.z)
            
        q = (4.0 / viewing_conditions.c) * math.sqrt(j / 100.0) * \
            (viewing_conditions.aw + 4.0) * viewing_conditions.fl_root
            
        hue_prime = hue + 360.0 if hue < 20.14 else hue
        
        e_hue = 0.25 * (math.cos((hue_prime * math.pi) / 180.0 + 2.0) + 3.8)
        
        p1 = (50000.0 / 13.0) * e_hue * viewing_conditions.nc * viewing_conditions.ncb
        
        t = (p1 * math.sqrt(a * a + b * b)) / (u + 0.305)
        alpha = math.pow(t, 0.9) * \
            math.pow(1.64 - math.pow(0.29, viewing_conditions.n), 0.73)
            
        c = alpha * math.sqrt(j / 100.0)
        
        m = c * viewing_conditions.fl_root
        
        s = 50.0 * \
            math.sqrt((alpha * viewing_conditions.c) / (viewing_conditions.aw + 4.0))
            
        jstar = ((1.0 + 100.0 * 0.007) * j) / (1.0 + 0.007 * j)
        mstar = (1.0 / 0.0228) * math.log(1.0 + 0.0228 * m)
        astar = mstar * math.cos(hue_radians)
        bstar = mstar * math.sin(hue_radians)
        
        return cls(hue, c, j, q, m, s, jstar, astar, bstar)
    
    @classmethod
    def from_jch(
        cls,
        j: float,
        c: float,
        h: float
    ) -> 'Cam16':
        """
        Creates a CAM16 color from J, C, and h in default viewing conditions.

        Args:
            j: Lightness
            c: Chroma
            h: Hue

        Returns:
            CAM16 color
        """
        return cls.from_jch_in_viewing_conditions(j, c, h, ViewingConditions.DEFAULT)
    
    @classmethod
    def from_jch_in_viewing_conditions(
        cls,
        j: float,
        c: float,
        h: float,
        viewing_conditions: ViewingConditions
    ) -> 'Cam16':
        """
        Creates a CAM16 color from J, C, and h in the given viewing conditions.

        Args:
            j: Lightness
            c: Chroma
            h: Hue
            viewing_conditions: Information about the environment where the color
                will be viewed.

        Returns:
            CAM16 color
        """
        q = (4.0 / viewing_conditions.c) * math.sqrt(j / 100.0) * \
            (viewing_conditions.aw + 4.0) * viewing_conditions.fl_root
            
        m = c * viewing_conditions.fl_root
        
        alpha = c / math.sqrt(j / 100.0)
        
        s = 50.0 * \
            math.sqrt((alpha * viewing_conditions.c) / (viewing_conditions.aw + 4.0))
            
        hue_radians = (h * math.pi) / 180.0
        
        jstar = ((1.0 + 100.0 * 0.007) * j) / (1.0 + 0.007 * j)
        mstar = (1.0 / 0.0228) * math.log(1.0 + 0.0228 * m)
        astar = mstar * math.cos(hue_radians)
        bstar = mstar * math.sin(hue_radians)
        
        return cls(h, c, j, q, m, s, jstar, astar, bstar)
    
    @classmethod
    def from_ucs(
        cls,
        jstar: float,
        astar: float,
        bstar: float
    ) -> 'Cam16':
        """
        Creates a CAM16 color from CAM16-UCS coordinates in default viewing conditions.

        Args:
            jstar: CAM16-UCS lightness.
            astar: CAM16-UCS a dimension. Like a* in L*a*b*, it is a Cartesian
                coordinate on the Y axis.
            bstar: CAM16-UCS b dimension. Like a* in L*a*b*, it is a Cartesian
                coordinate on the X axis.

        Returns:
            CAM16 color
        """
        return cls.from_ucs_in_viewing_conditions(
            jstar, astar, bstar, ViewingConditions.DEFAULT)
    
    @classmethod
    def from_ucs_in_viewing_conditions(
        cls,
        jstar: float,
        astar: float,
        bstar: float,
        viewing_conditions: ViewingConditions
    ) -> 'Cam16':
        """
        Creates a CAM16 color from CAM16-UCS coordinates in the given viewing conditions.

        Args:
            jstar: CAM16-UCS lightness.
            astar: CAM16-UCS a dimension. Like a* in L*a*b*, it is a Cartesian
                coordinate on the Y axis.
            bstar: CAM16-UCS b dimension. Like a* in L*a*b*, it is a Cartesian
                coordinate on the X axis.
            viewing_conditions: Information about the environment where the color
                will be viewed.

        Returns:
            CAM16 color
        """
        a = astar
        b = bstar
        m = math.sqrt(a * a + b * b)
        M = (math.exp(m * 0.0228) - 1.0) / 0.0228
        c = M / viewing_conditions.fl_root
        h = math.atan2(b, a) * (180.0 / math.pi)
        if h < 0.0:
            h += 360.0
        j = jstar / (1 - (jstar - 100) * 0.007)
        
        return cls.from_jch_in_viewing_conditions(j, c, h, viewing_conditions)
    
    def to_int(self) -> int:
        """
        Converts this color to ARGB format.

        Returns:
            ARGB representation of color, assuming the color was viewed in default
            viewing conditions, which are near-identical to the default viewing
            conditions for sRGB.
        """
        return self.viewed(ViewingConditions.DEFAULT)
    
    def viewed(self, viewing_conditions: ViewingConditions) -> int:
        """
        Converts this color to ARGB format.

        Args:
            viewing_conditions: Information about the environment where the color
                will be viewed.

        Returns:
            ARGB representation of color
        """
        alpha = 0.0 if self.chroma == 0.0 or self.j == 0.0 else \
            self.chroma / math.sqrt(self.j / 100.0)
        
        t = math.pow(
            alpha / math.pow(1.64 - math.pow(0.29, viewing_conditions.n), 0.73),
            1.0 / 0.9)
            
        h_rad = (self.hue * math.pi) / 180.0
        
        e_hue = 0.25 * (math.cos(h_rad + 2.0) + 3.8)
        
        ac = viewing_conditions.aw * \
            math.pow(
                self.j / 100.0, 1.0 / viewing_conditions.c / viewing_conditions.z)
                
        p1 = e_hue * (50000.0 / 13.0) * viewing_conditions.nc * viewing_conditions.ncb
        
        p2 = ac / viewing_conditions.nbb
        
        h_sin = math.sin(h_rad)
        h_cos = math.cos(h_rad)
        
        gamma = (23.0 * (p2 + 0.305) * t) / \
            (23.0 * p1 + 11.0 * t * h_cos + 108.0 * t * h_sin)
            
        a = gamma * h_cos
        b = gamma * h_sin
        
        r_a = (460.0 * p2 + 451.0 * a + 288.0 * b) / 1403.0
        g_a = (460.0 * p2 - 891.0 * a - 261.0 * b) / 1403.0
        b_a = (460.0 * p2 - 220.0 * a - 6300.0 * b) / 1403.0
        
        r_c_base = max(0, (27.13 * abs(r_a)) / (400.0 - abs(r_a)))
        r_c = math_utils.signum(r_a) * (100.0 / viewing_conditions.fl) * \
            math.pow(r_c_base, 1.0 / 0.42)
            
        g_c_base = max(0, (27.13 * abs(g_a)) / (400.0 - abs(g_a)))
        g_c = math_utils.signum(g_a) * (100.0 / viewing_conditions.fl) * \
            math.pow(g_c_base, 1.0 / 0.42)
            
        b_c_base = max(0, (27.13 * abs(b_a)) / (400.0 - abs(b_a)))
        b_c = math_utils.signum(b_a) * (100.0 / viewing_conditions.fl) * \
            math.pow(b_c_base, 1.0 / 0.42)
            
        r_f = r_c / viewing_conditions.rgb_d[0]
        g_f = g_c / viewing_conditions.rgb_d[1]
        b_f = b_c / viewing_conditions.rgb_d[2]
        
        x = 1.86206786 * r_f - 1.01125463 * g_f + 0.14918677 * b_f
        y = 0.38752654 * r_f + 0.62144744 * g_f - 0.00897398 * b_f
        z = -0.01584150 * r_f - 0.03412294 * g_f + 1.04996444 * b_f
        
        argb = color_utils.argb_from_xyz(x, y, z)
        return argb
    
    @classmethod
    def from_xyz_in_viewing_conditions(
        cls,
        x: float,
        y: float,
        z: float,
        viewing_conditions: ViewingConditions
    ) -> 'Cam16':
        """
        Creates a CAM16 color from XYZ coordinates in the given viewing conditions.

        Args:
            x: X coordinate
            y: Y coordinate
            z: Z coordinate
            viewing_conditions: Information about the environment where the color
                was observed.

        Returns:
            CAM16 color
        """
        # Transform XYZ to 'cone'/'rgb' responses
        r_c = 0.401288 * x + 0.650173 * y - 0.051461 * z
        g_c = -0.250268 * x + 1.204414 * y + 0.045854 * z
        b_c = -0.002079 * x + 0.048952 * y + 0.953127 * z
        
        # Discount illuminant
        r_d = viewing_conditions.rgb_d[0] * r_c
        g_d = viewing_conditions.rgb_d[1] * g_c
        b_d = viewing_conditions.rgb_d[2] * b_c
        
        # Chromatic adaptation
        r_af = math.pow(viewing_conditions.fl * abs(r_d) / 100.0, 0.42)
        g_af = math.pow(viewing_conditions.fl * abs(g_d) / 100.0, 0.42)
        b_af = math.pow(viewing_conditions.fl * abs(b_d) / 100.0, 0.42)
        
        r_a = math_utils.signum(r_d) * 400.0 * r_af / (r_af + 27.13)
        g_a = math_utils.signum(g_d) * 400.0 * g_af / (g_af + 27.13)
        b_a = math_utils.signum(b_d) * 400.0 * b_af / (b_af + 27.13)
        
        # redness-greenness
        a = (11.0 * r_a + -12.0 * g_a + b_a) / 11.0
        # yellowness-blueness
        b = (r_a + g_a - 2.0 * b_a) / 9.0
        
        # auxiliary components
        u = (20.0 * r_a + 20.0 * g_a + 21.0 * b_a) / 20.0
        p2 = (40.0 * r_a + 20.0 * g_a + b_a) / 20.0
        
        # hue
        atan2 = math.atan2(b, a)
        atan_degrees = atan2 * 180.0 / math.pi
        
        hue = atan_degrees if atan_degrees >= 0 else atan_degrees + 360.0
        hue = hue if hue < 360 else hue - 360.0
        
        hue_radians = hue * math.pi / 180.0
        
        # achromatic response to color
        ac = p2 * viewing_conditions.nbb
        
        # CAM16 lightness and brightness
        J = 100.0 * \
            math.pow(
                ac / viewing_conditions.aw,
                viewing_conditions.c * viewing_conditions.z)
                
        Q = (4.0 / viewing_conditions.c) * math.sqrt(J / 100.0) * \
            (viewing_conditions.aw + 4.0) * (viewing_conditions.fl_root)
            
        hue_prime = hue + 360.0 if hue < 20.14 else hue
        
        e_hue = (1.0 / 4.0) * (math.cos(hue_prime * math.pi / 180.0 + 2.0) + 3.8)
        
        p1 = 50000.0 / 13.0 * e_hue * viewing_conditions.nc * viewing_conditions.ncb
        
        t = p1 * math.sqrt(a * a + b * b) / (u + 0.305)
        
        alpha = math.pow(t, 0.9) * \
            math.pow(1.64 - math.pow(0.29, viewing_conditions.n), 0.73)
            
        # CAM16 chroma, colorfulness, saturation
        C = alpha * math.sqrt(J / 100.0)
        
        M = C * viewing_conditions.fl_root
        
        s = 50.0 * \
            math.sqrt((alpha * viewing_conditions.c) / (viewing_conditions.aw + 4.0))
            
        # CAM16-UCS components
        jstar = (1.0 + 100.0 * 0.007) * J / (1.0 + 0.007 * J)
        mstar = math.log(1.0 + 0.0228 * M) / 0.0228
        astar = mstar * math.cos(hue_radians)
        bstar = mstar * math.sin(hue_radians)
        
        return cls(hue, C, J, Q, M, s, jstar, astar, bstar)
    
    def xyz_in_viewing_conditions(self, viewing_conditions: ViewingConditions) -> list[float]:
        """
        Converts a CAM16 color to XYZ coordinates in the given viewing conditions.

        Args:
            viewing_conditions: Information about the environment where the color
                will be viewed.

        Returns:
            XYZ coordinates
        """
        alpha = 0.0 if self.chroma == 0.0 or self.j == 0.0 else \
            self.chroma / math.sqrt(self.j / 100.0)
        
        t = math.pow(
            alpha / math.pow(1.64 - math.pow(0.29, viewing_conditions.n), 0.73),
            1.0 / 0.9)
            
        h_rad = self.hue * math.pi / 180.0
        
        e_hue = 0.25 * (math.cos(h_rad + 2.0) + 3.8)
        
        ac = viewing_conditions.aw * \
            math.pow(
                self.j / 100.0, 1.0 / viewing_conditions.c / viewing_conditions.z)
                
        p1 = e_hue * (50000.0 / 13.0) * viewing_conditions.nc * viewing_conditions.ncb
        
        p2 = (ac / viewing_conditions.nbb)
        
        h_sin = math.sin(h_rad)
        h_cos = math.cos(h_rad)
        
        gamma = 23.0 * (p2 + 0.305) * t / \
            (23.0 * p1 + 11 * t * h_cos + 108.0 * t * h_sin)
            
        a = gamma * h_cos
        b = gamma * h_sin
        
        r_a = (460.0 * p2 + 451.0 * a + 288.0 * b) / 1403.0
        g_a = (460.0 * p2 - 891.0 * a - 261.0 * b) / 1403.0
        b_a = (460.0 * p2 - 220.0 * a - 6300.0 * b) / 1403.0
        
        r_c_base = max(0, (27.13 * abs(r_a)) / (400.0 - abs(r_a)))
        r_c = math_utils.signum(r_a) * (100.0 / viewing_conditions.fl) * \
            math.pow(r_c_base, 1.0 / 0.42)
            
        g_c_base = max(0, (27.13 * abs(g_a)) / (400.0 - abs(g_a)))
        g_c = math_utils.signum(g_a) * (100.0 / viewing_conditions.fl) * \
            math.pow(g_c_base, 1.0 / 0.42)
            
        b_c_base = max(0, (27.13 * abs(b_a)) / (400.0 - abs(b_a)))
        b_c = math_utils.signum(b_a) * (100.0 / viewing_conditions.fl) * \
            math.pow(b_c_base, 1.0 / 0.42)
            
        r_f = r_c / viewing_conditions.rgb_d[0]
        g_f = g_c / viewing_conditions.rgb_d[1]
        b_f = b_c / viewing_conditions.rgb_d[2]
        
        x = 1.86206786 * r_f - 1.01125463 * g_f + 0.14918677 * b_f
        y = 0.38752654 * r_f + 0.62144744 * g_f - 0.00897398 * b_f
        z = -0.01584150 * r_f - 0.03412294 * g_f + 1.04996444 * b_f
        
        return [x, y, z]